"""

                        This is the process module

This module reads BioSANS topology file, grab the components or species,
rate  constants, stoichiometric  matrix,  propensity   vector, algebraic
rules, conditional statements, and othe types of definitions into a dic-
tionary.  This module  calls the process_hub  module that distribute the
tasks to other modules.

The functions in this module are as follows; #

1. eval_dict
2. tofloat
3. is_number
4. process


"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

# import warnings
from tkinter import messagebox as message_upon_error
# import re
# warnings.filterwarnings('ignore')

# import time
import numpy as np

from BioSANS2020.prepcodes.processes_hub import process_hub
# from BioSANS2020.myglobal import mglobals as globals2
# from BioSANS2020.myglobal import proc_global
from BioSANS2020.math_functs.sbml_math import SBML_FUNCT_DICT


def eval_dict(to_eval, loc_dict):
    """This function takes a string expression and return the evaluated
    expression using SBML_FUNCT_DICT and the locals() dictionary where
    eval_dict is called.

    Args:
        to_eval (str): the expression to evaluate
        loc_dict (dict): local dictionary from the calling function

    Returns:
        multitype: result of eval command
    """
    return eval(to_eval, loc_dict, SBML_FUNCT_DICT)


def tofloat(val, loc_dict):
    """This function attempts to convert the input val into float

    Args:
        val (str): the expression to evaluate
        loc_dict (dict): local dictionary from the calling function

    Returns:
        float: float equivalent of val
    """
    try:
        return float(val)
    except:
        return float(eval_dict(val, loc_dict))


def is_number(xvar):
    """This function checks if a string xvar is float

    Args:
        xvar (str): input string expression or number

    Returns:
        bool: True if xvar cal be converted to float otherwise False
    """
    try:
        float(xvar)
        return True
    except:
        return False


def process(
        rfile="Reactions",
        miter=1,
        conc_unit="molecules",
        v_volms=1.0e-20,
        tend=1,
        del_coef=10,
        normalize=False,
        logx=False,
        logy=False,
        method="CLE",
        tlen=1000,
        mix_plot=True,
        save=True,
        out_fname="",
        plot_show=True,
        time_unit="time (sec)",
        vary="",
        vary2="",
        mult_proc=False,
        implicit=False,
        items=None,
        exp_data_file=None,
        c_input=None
):
    """[summary]

    Args:
        rfile (str): file name of BioSANS topology file.
        miter (int, optional): Number of iteration or trajectory samples
            for stochastic integration
        conc_unit (bool, optional): "mole","molar", or "molecules" - the
            unit used in any amount in topology file. Defaults to
            "molecules".
        v_volms (float, optional): the volume of compartment used in the
            simulation. Defaults to 1.0e-20.
        tend (float): trajectory simulation end time. Defaults to 1.
        del_coef (float, optional): factor for modifying time steps used
            in the integration/propagation of ODE. Defaults to 10.
        normalize (bool, optional): True  will  be normalized the y axis
            based on max value . Defaults to False.
        logx (bool, optional): If True, the x-axis will be in log scale.
            Defaults to False.
        logy (bool, optional): if True, the y-axis will be in log scale.
            Defaults to False.
        method (str, optional): Defaults to "CLE". Any of the option in
            the list of available method keywords is listed below;

            Stochastic (refer to section 10.2.4)

            1.	"CLE"            - Molecules(micro), tau-adaptive
            2.	"CLE2"           - Molecules(micro), cle-fixIntvl
            3.	"Gillespie_"     - Molecules(micro), Direct method
            4.	"Tau-leaping"    - Molecules(micro),
                                   Not swapping with Gillespie
            5.	"Tau-leaping2"   - Molecules(micro),
                                   Swapping with Gillespie
            6.	"Sim-TauLeap"    - Molecules(micro), Simplified,
                                   Swapping with Gillespie

            Deterministic (refer to section 10.2.1)

            7.	"Euler-1"        - Molecules(micro), tau-adaptive-1
            8.	"Euler-2"        - Molar (macro), tau-adaptive-1
            9.	"Euler-3"        - Mole (macro), tau-adaptive-1
            10.	"Euler2-1"	     - Molecules(micro), tau-adaptive-2
            11.	"Euler2-2"       - Molar (macro), tau-adaptive-2
            12.	"Euler2-3"       - Mole (macro), tau-adaptive-2
            13.	"ODE-1"          - Molecules(micro),
                                   using ode_int from scipy
            14.	"ODE-2"          - Molar(macro),
                                   using ode_int from scipy
            15.	"ODE-3"          - Mole(macro), using ode_int from scipy
            16.	"rk4-1"          - Molecules(micro), fix-interval
            17.	"rk4-2"          - Molar(macro), fix-interval
            18.	"rk4-3"          - Mole(macro), fix-interval
            19.	"rk4-1a"         - Molecules(micro), tau-adaptive
            20.	"rk4-2a"         - Molar(macro), tau-adaptive
            21.	"rk4-3a"         - Mole(macro), tau-adaptive

            Linear Noise Approximation (refer to 10.1.2 & 10.2.2)

            22.	"LNA"             - Numeric, values
            23.	"LNA-vs"          - Symbolic, values, Macroscopic
            24.	"LNA-ks"          - Symbolic, f(ks), Macroscopic
            25.	"LNA-xo"          - Symbolic, f(xo), Macroscopic
            26.	"LNA2"            - Symbolic, f(xo,ks), Microscopic
            27.	"LNA3"            - Symbolic, f(xo,ks), Macroscopic
            28.	"LNA(t)"          - COV-time-dependent, Macroscopic
            29.	"LNA2(t)"         - FF-time-dependent, Macroscopic

            Network Localization (refer to 10.1.3)

            30.	"NetLoc1"         - Symbolic, Macroscopic
            31.	"NetLoc2"         - Numeric, Macroscopic

            Parameter estimation (refer to 10.2.3)

            32.	"k_est1"          - MCEM, Macroscopic
            33.	"k_est2"          - MCEM, Microscopic
            34.	"k_est3"          - NM-Diff. Evol., Macroscopic
            35.	"k_est4"          - NM-Diff. Evol., Microscopic
            36.	"k_est5"          - Parameter slider/scanner
            37.	"k_est6"          - Nelder-Mead (NM), Macroscopic
            38.	"k_est7"          - Nelder-Mead (NM), Microscopic
            39.	"k_est8"          - Powell, Macroscopic
            40.	"k_est9"          - Powell, Microscopic
            41.	"k_est10"         - L-BFGS-B, Macroscopic
            42.	"k_est11"         - L-BFGS-B, Microscopic

            Symbolic/Analytical expression of species (refer to 10.1.1)

            43.	"Analyt"          - Pure Symbolic :f(t,xo,k)
            44.	"Analyt-ftx"      - Semi-Symbolic :f(t,xo)
            45.	"SAnalyt"         - Semi-Symbolic :f(t)
            46.	"SAnalyt-ftk"     - Semi-Symbolic :f(t,k)
            47.	"Analyt2"         - Creates commands for wxmaxima

        tlen (int, optional): number of integration steps reported in
            the final result. Defaults to 1000.
        mix_plot (bool, optional): If True, all species are plotted in
            one plot/figure. Defaults to True.
        save (bool, optional): If True, the resulting trajectory in the
            simulation will be saved as a file. Defaults to True.
        out_fname (str, optional): output filename. Defaults to "".
        plot_show (bool, optional): If True, an image of the plots will
            be created in the directory of the topology file.
        time_unit (str, optional): Defaults to "time (sec)".
        vary (str, optional): Varying initial concentration.
            Defaults to "".
        vary2 (str, optional): [description]. Defaults to "".
        mult_proc (bool, optional): If True, trajectories will be propa-
            gated on parallel. Defaults to False.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        items (tuple, optional): (canvas, scroll_x, scroll_y).
            Defaults to None.
        exp_data_file ([type], optional): Experimental data file contai-
            ning True or accepted trajectories. Defaults to None.
        c_input (dict, optional): [description]. Defaults to {}.

    Returns:
        list: list of simulated trajecotry.
            data[j][0] - time for trajectory j
            data[j][1][:, i] - trajectories of each component i
    """

    avogadros_num = 6.022e+23

    # globals2.MODIFIED = {}
    # globals2.PROP_MODIFIED = {}

    # for func in globals2.EXEC_FUNCTIONS:
    #     exec(func, globals(), SBML_FUNCT_DICT)

    if conc_unit == "molar":
        if method in ["ODE-1", "rk4-1", "rk4-1a", "Euler-1",
                      "Euler2-1", "CLE", "CLE2", "Tau-leaping",
                      "Tau-leaping2", "Sim-TauLeap", "Gillespie_"]:
            avgdrs_times_vol = avogadros_num * v_volms
            factor_stch = 2
        elif method in ["ODE-3", "rk4-3", "rk4-3a",
                        "Euler-3", "Euler2-3"]:
            avgdrs_times_vol = v_volms
            factor_stch = 1
        else:
            avgdrs_times_vol = 1
            factor_stch = 1
    elif conc_unit == "molecules":
        if method in ["ODE-2", "rk4-2", "rk4-2a",
                      "Euler-2", "Euler2-2"]:
            avgdrs_times_vol = 1 / avogadros_num * v_volms
            factor_stch = 1 / 2
        elif method in ["ODE-3", "rk4-3", "rk4-3a",
                        "Euler-3", "Euler2-3"]:
            avgdrs_times_vol = 1 / avogadros_num
            factor_stch = 1 / 2
        else:
            avgdrs_times_vol = 1
            factor_stch = 1
    elif conc_unit == "moles":
        if method in ["ODE-1", "rk4-1", "rk4-1a", "Euler-1", "Euler2-1",
                      "CLE", "CLE2", "Tau-leaping", "Tau-leaping2",
                      "Sim-TauLeap", "Gillespie_"]:
            avgdrs_times_vol = avogadros_num
            factor_stch = 2
        if method in ["ODE-2", "rk4-2", "rk4-2a",
                      "Euler-2", "Euler2-2"]:
            avgdrs_times_vol = 1 / v_volms
            factor_stch = 1
        else:
            avgdrs_times_vol = 1
            factor_stch = 1
    with open(rfile, "r") as file:
        rows = []
        conc = {}

        last = ""
        for row in file:
            row = row.strip()+" "
            if last == "Function_Definitions":
                if row.strip() != "" and row[0] != "#":
                    exec(row.strip(), locals(), SBML_FUNCT_DICT)
                elif row[0] == "#":
                    last = "#"
            elif last == "#":
                if row.strip() != "" and row[0] != "@":
                    rows.append(row)
                elif row.strip() != "" and row[0] != "#":
                    last = "@"
            elif last == "@":
                if row.strip() != "" and row[0] != "@":
                    cvar = row.split(",")
                    conc[cvar[0].strip()] = tofloat(cvar[1], locals())
                    # if len(cvar)>=3:
                    # cc = ",".join(cvar[2:])
                    # cc2 = cc.split(":")[0].replace("lambda","") \
                    #     .split(",")
                    # globals2.MODIFIED[cvar[0].strip()] = \
                    #     [cc2,eval_dict(cc)]
            elif row[0] == "#":
                last = "#"
                # gg = row.split(",")[1:]
                # try:
                # for xvar in gg:
                # xx = xvar.split("=")
                # globals2.SETTINGS[xx[0].strip()] = xx[1].strip()
                # except:
                # pass
            elif row[0] == "@":
                last = "@"
            elif row.strip().upper() == "FUNCTION_DEFINITIONS:":
                last = "Function_Definitions"

        file.close()

    ks_dict = {}
    r_dict = {}
    p_dict = {}
    sp_comp = {}
    rxn_rows = len(rows)
    for ih_ind in range(rxn_rows):
        r_dict[ih_ind] = {}
        p_dict[ih_ind] = {}
        col_row = rows[ih_ind].split(":::::")
        row = col_row[0].strip().split(",")
        # if len(col_row)>1:
        # krow = col_row[1].strip().split(":::")
        # if len(krow)==2:
        # cc2 = krow[0].split(":")[0].replace("lambda","").split(",")
        # cc3 = krow[1].split(":")[0].replace("lambda","").split(",")
        # globals2.PROP_MODIFIED["Prop_"+str(ih_ind)] = [
        #     (cc2,eval_dict(krow[0])),(cc3,eval_dict(krow[1]))]
        # else:
        # cc2 = krow[0].split(":")[0].replace("lambda","").split(",")
        # globals2.PROP_MODIFIED["Prop_"+str(ih_ind)] = \
        #     [(cc2,eval_dict(krow[0]))]

        if len(row) == 3:
            ks_dict[ih_ind] = [
                tofloat(row[1], locals()),
                tofloat(row[2], locals())]
        else:
            ks_dict[ih_ind] = [tofloat(row[1], locals())]
        col_var = row[0].split("<=>")
        if len(col_var) == 1:
            col_var = row[0].split("=>")

        sp_c = col_var[0]
        svar = sp_c.strip().split()
        if len(svar) > 1:
            last = 1
            for xvar in svar:
                if not is_number(xvar) and xvar != "+":
                    r_dict[ih_ind][xvar] = last
                    last = 1
                    if xvar in sp_comp:
                        sp_comp[xvar].add(ih_ind)
                    else:
                        sp_comp[xvar] = {ih_ind}
                elif is_number(xvar):
                    last = tofloat(xvar, locals())
        else:
            xvar = svar[0]
            r_dict[ih_ind][xvar] = 1
            if xvar in sp_comp:
                sp_comp[xvar].add(ih_ind)
            else:
                sp_comp[xvar] = {ih_ind}

        sp_c = col_var[1]
        svar = sp_c.strip().split()
        if len(svar) > 1:
            last = 1
            for xvar in svar:
                if (not is_number(xvar) or xvar.lower() == "e") \
                        and xvar != "+":
                    p_dict[ih_ind][xvar] = last
                    last = 1
                    if xvar in sp_comp:
                        sp_comp[xvar].add(ih_ind)
                    else:
                        sp_comp[xvar] = {ih_ind}
                elif is_number(xvar):
                    last = tofloat(xvar, locals())
        else:
            xvar = svar[0]
            p_dict[ih_ind][xvar] = 1
            if xvar in sp_comp:
                sp_comp[xvar].add(ih_ind)
            else:
                sp_comp[xvar] = {ih_ind}
    # print(sp_comp)

    stoch_var = []

    for sp_c in sp_comp:
        row = []
        for r_ind in range(rxn_rows):
            prod = p_dict[r_ind][sp_c] if sp_c in p_dict[r_ind] else 0
            rect = r_dict[r_ind][sp_c] if sp_c in r_dict[r_ind] else 0
            row.append(prod - rect)
            if len(ks_dict[r_ind]) == 2:
                row.append(-row[-1])
        stoch_var.append(row)

    stoch_var = np.array(stoch_var)

    ksn_dict = {}
    concn = {}
    try:
    # if True:
        for r_ind in range(rxn_rows):
            ksn_dict[r_ind] = [0] * 2
            if len(r_dict[r_ind]) == 1:
                for xvar in r_dict[r_ind]:
                    if r_dict[r_ind][xvar] == 0:
                        ksn_dict[r_ind][0] = ks_dict[r_ind][0] \
                            * avgdrs_times_vol
                    elif r_dict[r_ind][xvar] == 1:
                        ksn_dict[r_ind][0] = ks_dict[r_ind][0]
                    elif r_dict[r_ind][xvar] == 2:
                        ksn_dict[r_ind][0] = factor_stch * ks_dict[r_ind][0] \
                            / avgdrs_times_vol
                    else:
                        ksn_dict[r_ind][0] = ks_dict[r_ind][0]
                    concn[xvar] = conc[xvar] * avgdrs_times_vol

            elif len(r_dict[r_ind]) == 2:
                ksn_dict[r_ind][0] = ks_dict[r_ind][0] / avgdrs_times_vol
                for xvar in r_dict[r_ind]:
                    concn[xvar] = conc[xvar] * avgdrs_times_vol

            if len(p_dict[r_ind]) == 1:
                for xvar in p_dict[r_ind]:
                    if p_dict[r_ind][xvar] == 0:
                        if len(ks_dict[r_ind]) == 2:
                            ksn_dict[r_ind][1] = ks_dict[r_ind][1] \
                                * avgdrs_times_vol
                        else:
                            ksn_dict[r_ind] = [ksn_dict[r_ind][0]]
                    elif p_dict[r_ind][xvar] == 1:
                        if len(ks_dict[r_ind]) == 2:
                            ksn_dict[r_ind][1] = ks_dict[r_ind][1]
                        else:
                            ksn_dict[r_ind] = [ksn_dict[r_ind][0]]
                    elif p_dict[r_ind][xvar] == 2:
                        if len(ks_dict[r_ind]) == 2:
                            ksn_dict[r_ind][1] = factor_stch \
                                * ks_dict[r_ind][1] / avgdrs_times_vol
                        else:
                            ksn_dict[r_ind] = [ksn_dict[r_ind][0]]
                    else:
                        if len(ks_dict[r_ind]) == 2:
                            ksn_dict[r_ind][1] = ks_dict[r_ind][1]
                        else:
                            ksn_dict[r_ind] = [ksn_dict[r_ind][0]]
                    concn[xvar] = conc[xvar] * avgdrs_times_vol
            elif len(p_dict[r_ind]) == 2:
                if len(ks_dict[r_ind]) == 2:
                    ksn_dict[r_ind][1] = ks_dict[r_ind][1] / avgdrs_times_vol
                else:
                    ksn_dict[r_ind] = [ksn_dict[r_ind][0]]
                for xvar in p_dict[r_ind]:
                    concn[xvar] = conc[xvar] * avgdrs_times_vol

        for xvar in conc:
            if xvar not in concn:
                concn[xvar] = conc[xvar]

        # for xvar in c_input:
            # concn[xvar] = c_input[xvar]

        # t_o = time.time()
        tvar = np.linspace(0, tend, int(tlen + 1))
        data = process_hub(
            tvar, sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
            v_volms, miter, logx, logy, del_coef, normalize, method,
            mix_plot, save, out_fname, plot_show, time_unit, vary,
            mult_proc, items, vary2, implicit, rfile, exp_data_file)
        # print(time.time()-t_o,"Process time")

        return data
    except Exception as error:
        print(c_input)
        print(error)
        message_upon_error.showerror(
            "showinfo",
            "Check your topology files for missing species \
                in reaction and concentration tag : " + str(error))



"""

                  This module is the RunBioSANS module

This can be used to run BioSANS and BioSSL byb running the following.

BioSSL()
BioSANS()


"""

from os import system
from subprocess import Popen
from sys import executable


def BioSANS():
    """This function launched BioSANS"""
    Popen([str(executable), "-m", "BioSANS2020.BioSANS"])


def BioSSL():
    """This function launched BioSSL"""
    system(str(executable) + " -m BioSANS2020.BioSSL")


def biosans():
    """This function launched BioSANS"""
    Popen([str(executable), "-m", "BioSANS2020.BioSANS"])


def biossl():
    """This function launched BioSSL"""
    system(str(executable) + " -m BioSANS2020.BioSSL")



# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

"""
                       This is the ssl_calls module

This module interacts with BioSSL.py by fulfilling its requests to pro-
vide a console interface that supports some features of BioSANS.

The following is the list of function for this module:

1. load_data_traj
2. calc_average_conc_at_tend
3. calc_covariance
4. calc_covariance_per_traj
5. calc_covariance_bootsrap
6. prob_density_calc_wtime
7. prob_density_calc_tslice
8. prob_density_calc

"""

import numpy as np
import matplotlib.pyplot as plt
from BioSANS2020.analysis.numeric.sample_points import sample_points


def load_data_traj(file_name):
    """This function loads trajectory data from a tab delimited file.
    The first column in the file is time, the remaining columns are
    species or components. All sampled trajectories are concatenated
    in the file.
    Args:
        file_name : name of trajectory file generated in BioSANS
                    simulations (either deterministic or stochastic).
    Return:
        current_data : two dimensional array of [data, slabels].
                       slabels are the names in the header of file_name.
                       data is a list of trajectory data w/o header

    """
    with open(file_name, "r") as ffile:
        data = []
        ddvar = []
        row1 = str(ffile.readline()).strip()
        slabels = row1.split("\t")[1:]
        for row in ffile:
            cols = [float(x) for x in row.split("\t")]
            if cols[0] == 0.0 and len(ddvar) > 0:
                data.append(np.array(ddvar))
                ddvar = []
            ddvar.append(cols)
        data.append(np.array(ddvar))
    current_data = (data, slabels)
    return current_data


def calc_average_conc_at_tend(edata, points=100):
    """This function calculates the average or mean of edata[0] using
    the last number of points in the trajectory. If the simulation is
    long enought, this is the steady state mean concentration.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        points : number of data points to slice at end of trajectory
    """
    data, slabels = edata
    nlen = len(data)
    ddm = 0
    for i in range(1, nlen):
        ddvar = data[i][-points:, 1:]
        ddm = ddm + ddvar

    ddm = np.mean(ddm / (nlen - 1), 0)
    dlen = len(slabels)
    print("\nSpecies Concentration\n")
    for i in range(dlen):
        if str(ddm[i]) not in {"None", "nan", "0.0"}:
            print(slabels[i].ljust(50) + " = " + str(ddm[i]))


def calc_covariance(edata, points=100):
    """This function calculates the covariance of edata[0] and prints
    the result in a console.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        points : last number of points considered in covariance
                 calculation from -points to the end of array
                 or equivalent to [-points:] slice.
    """
    data, slabels = edata
    nlen = len(data)
    ddm = 0
    for i in range(1, nlen):
        ddvar = data[i][-points:, 1:]
        ddm = ddm + ddvar

    ddm = np.mean(ddm / (nlen - 1), 0)

    vvar = 0
    dlen = len(slabels)
    for k in range(1, nlen):
        ddvar = data[k][-points:, 1:] - ddm
        ddd = np.zeros((dlen, dlen))
        for i in range(dlen):
            for j in range(dlen):
                ddd[i, j] = np.mean(ddvar[:, i] * ddvar[:, j])
        vvar = vvar + ddd
    vvar = vvar / (nlen - 1)

    print("covariance\n")
    for i in range(dlen):
        for j in range(i, dlen):
            if str(vvar[i, j]) not in {"None", "nan", "0.0"}:
                label = slabels[i] + " and " + slabels[j]
                print(label.ljust(50) + " = " + str(vvar[i, j]))

    vvv = np.zeros((dlen, dlen))
    print("\ncorrelation\n")
    for i in range(dlen):
        for j in range(i, dlen):
            if str(vvar[i, j]) not in {"None", "nan", "0.0"}:
                vvv[i, j] = vvar[i, j] / np.sqrt(vvar[i, i] * vvar[j, j])
                label = slabels[i] + " and " + slabels[j]
                print(label.ljust(50) + " = " + str(vvv[i, j]))

    ffvar = 0
    cvcaps = 0
    for i in range(nlen):
        ddvar = 0
        ddvar = ddvar + (data[i][-points:, 1:] - ddm)**2
        ffvar = ffvar + np.mean(ddvar, 0) / ddm
        cvcaps = cvcaps + np.mean(np.sqrt(ddvar), 0) / ddm
    ffvar = ffvar / (nlen - 1)
    cvcaps = cvcaps / (nlen - 1)
    print("\nFano Factor\n")
    for i in range(dlen):
        if str(ffvar[i]) not in {"None", "nan", "0.0"}:
            print(slabels[i].ljust(50) + " = " + str(ffvar[i]))

    print("\nCoefficient of variation\n")
    for i in range(dlen):
        if str(cvcaps[i]) not in {"None", "nan", "0.0"}:
            print(slabels[i].ljust(50) + " = " + str(cvcaps[i]))


def calc_covariance_per_traj(edata, points=100, fname="", mname=""):
    """This function calculates the covariance of edata[0] per rows,
    prints the result in a console, and plots data into image.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        points : last number of points considered in covariance
         calculation from -points to the end of array ([-points:] slice)
        fname : prepended name to plots fname_mname*
        mname : prepended name to plots fname_mname*
    """
    data, slabels = edata
    nlen = len(data)
    ddm = 0
    for i in range(1, nlen):
        ddvar = data[i][-points:, 1:]
        ddm = ddm + ddvar

    ddm = np.mean(ddm / (nlen - 1), 0)

    vvar = []
    ffvar = []
    cvlcase = []
    dlen = len(slabels)
    for k in range(1, nlen):
        ddvar = data[k][-points:, 1:] - ddm
        ddd = np.zeros((dlen, dlen))
        fff = np.zeros(dlen)
        cvv = np.zeros(dlen)
        for i in range(dlen):
            for j in range(dlen):
                ddd[i, j] = np.mean(ddvar[:, i] * ddvar[:, j])
            fff[i] = ddd[i, i] / ddm[i]
            cvv[i] = np.sqrt(ddd[i, i]) / ddm[i]
        vvar.append(ddd)
        ffvar.append(fff)
        cvlcase.append(cvv)
    vvar = np.array(vvar)
    ffvar = np.array(ffvar)
    cvlcase = np.array(cvlcase)
    mvv = np.mean(vvar, axis=0)
    svv = np.std(vvar, axis=0)
    mff = np.mean(ffvar, axis=0)
    sff = np.std(ffvar, axis=0)
    mcv = np.mean(cvlcase, axis=0)
    scv = np.std(cvlcase, axis=0)
    print("\ncovariance\n")
    for i in range(dlen):
        for j in range(i, dlen):
            if str(mvv[i, j]) not in {"None", "nan", "0.0"}:
                label = slabels[i] + " and " + slabels[j]
                print(label.ljust(50) + " = " +
                      str(mvv[i, j]), "std =", svv[i, j])

    plt.figure(figsize=(9.68, 3.95))
    plt.xlabel("covariance(P)")
    plt.ylabel("prob")
    plt.hist(vvar[:, dlen - 1, dlen - 1], bins=50,
             density=True, orientation='vertical')
    plt.axvline(mvv[dlen - 1, dlen - 1])
    plt.tight_layout()
    plt.savefig(fname + "_" + mname + "_covariance(P).jpg")
    plt.close()

    print("\nFano Factor\n")
    for i in range(dlen):
        if str(mff[i]) not in {"None", "nan", "0.0"}:
            print(slabels[i].ljust(50) + " = " + str(mff[i]), "std =", sff[i])

    plt.figure(figsize=(9.68, 3.95))
    plt.xlabel("Fano factor(P)")
    plt.ylabel("prob")
    plt.hist(ffvar[:, dlen - 1], bins=50, density=True, orientation='vertical')
    plt.axvline(mff[dlen - 1])
    plt.tight_layout()
    plt.savefig(fname + "_" + mname + "_Fano factor(P).jpg")
    plt.close()

    print("\nCoefficient of variation\n")
    for i in range(dlen):
        if str(mcv[i]) not in {"None", "nan", "0.0"}:
            print(slabels[i].ljust(50) + " = " + str(mcv[i]), "std =", scv[i])

    plt.figure(figsize=(9.68, 3.95))
    plt.xlabel("Coeff. of Var.(P)")
    plt.ylabel("prob")
    plt.hist(cvlcase[:, dlen - 1], bins=50, density=True,
             orientation='vertical')
    plt.axvline(mcv[dlen - 1])
    plt.tight_layout()
    plt.savefig(fname + "_" + mname + "_CoeffOfVar(P).jpg")
    plt.close()


def calc_covariance_bootsrap(
        edata, points=100, msamp=1000, fname="", mname=""):
    """This function calculates the covariance of edata[0] with sampling
    , prints the result in a console, and plots data into image.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        points : last number of points considered in covariance
                 calculation from -points to the end of array
                 or equivalent to [-points:] slice.
        msamp : number of randomly chosen trajectories
        fname : prepended name to plots fname_mname*
        mname : prepended name to plots fname_mname*
    """
    data, slabels = edata
    nlen = len(data)
    bmvv = []
    bsvv = []
    bmff = []
    bsff = []
    bmcv = []
    bscv = []
    mlist = list(range(1, nlen))
    np.random.shuffle(mlist)
    for _ in range(1000):
        ddm = 0
        curr_list = np.random.choice(mlist, msamp)
        for i in curr_list:
            ddvar = data[i][-points:, 1:]
            ddm = ddm + ddvar

        ddm = np.mean(ddm / msamp, 0)

        vvar = []
        ffvar = []
        cvlcase = []
        dlen = len(slabels)
        for k in curr_list:
            ddvar = data[k][-points:, 1:] - ddm
            ddd = np.zeros((dlen, dlen))
            fff = np.zeros(dlen)
            cvv = np.zeros(dlen)
            for i in range(dlen):
                for j in range(dlen):
                    ddd[i, j] = np.mean(ddvar[:, i] * ddvar[:, j])
                fff[i] = ddd[i, i] / ddm[i]
                cvv[i] = np.sqrt(ddd[i, i]) / ddm[i]
            vvar.append(ddd)
            ffvar.append(fff)
            cvlcase.append(cvv)
        vvar = np.array(vvar)
        ffvar = np.array(ffvar)
        cvlcase = np.array(cvlcase)
        mvv = np.mean(vvar, axis=0)
        svv = np.std(vvar, axis=0)
        mff = np.mean(ffvar, axis=0)
        sff = np.std(ffvar, axis=0)
        mcv = np.mean(cvlcase, axis=0)
        scv = np.std(cvlcase, axis=0)
        bmvv.append(mvv)
        bsvv.append(svv)
        bmff.append(mff)
        bsff.append(sff)
        bmcv.append(mcv)
        bscv.append(scv)

    bmvv = np.array(bmvv)
    bsvv = np.array(bsvv)
    bmff = np.array(bmff)
    bsff = np.array(bsff)
    bmcv = np.array(bmcv)
    bscv = np.array(bscv)

    mvv = np.mean(bmvv, axis=0)
    svv = np.std(bmvv, axis=0)
    mff = np.mean(bmff, axis=0)
    sff = np.std(bmff, axis=0)
    mcv = np.mean(bmcv, axis=0)
    scv = np.std(bmcv, axis=0)

    print("\ncovariance\n")
    for i in range(dlen):
        for j in range(i, dlen):
            if str(mvv[i, j]) not in {"None", "nan", "0.0"}:
                label = slabels[i] + " and " + slabels[j]
                print(label.ljust(50) + " = " +
                      str(mvv[i, j]), "std =", svv[i, j])

    plt.figure(figsize=(9.68, 3.95))
    plt.xlabel("covariance(P)")
    plt.ylabel("prob")
    plt.hist(bmvv[:, dlen - 1, dlen - 1], bins=50,
             density=True, orientation='vertical')
    plt.axvline(mvv[dlen - 1, dlen - 1])
    plt.tight_layout()
    plt.savefig(fname + "_" + mname + "_covariance(P)_bootsrap.jpg")
    plt.close()

    print("\nFano Factor\n")
    for i in range(dlen):
        if str(mff[i]) not in {"None", "nan", "0.0"}:
            print(slabels[i].ljust(50) + " = " + str(mff[i]), "std =", sff[i])

    plt.figure(figsize=(9.68, 3.95))
    plt.xlabel("Fano factor(P)")
    plt.ylabel("prob")
    plt.hist(bmff[:, dlen - 1], bins=50, density=True, orientation='vertical')
    plt.axvline(mff[dlen - 1])
    plt.tight_layout()
    plt.savefig(fname + "_" + mname + "_Fano factor(P)_bootsrap.jpg")
    plt.close()

    print("\nCoefficient of Variation\n")
    for i in range(dlen):
        if str(mcv[i]) not in {"None", "nan", "0.0"}:
            print(slabels[i].ljust(50) + " = " + str(mcv[i]), "std =", scv[i])

    plt.figure(figsize=(9.68, 3.95))
    plt.xlabel("Coeff. of Var.")
    plt.ylabel("prob")
    plt.hist(bmcv[:, dlen - 1], bins=50, density=True, orientation='vertical')
    plt.axvline(mcv[dlen - 1])
    plt.tight_layout()
    plt.savefig(fname + "_" + mname + "_CoeffofVar(P)_bootsrap.jpg")
    plt.close()


def prob_density_calc_wtime(edata, fname, mname):
    """This function calculates the probability density of edata[0] and
    returns a plot of the probability density with time.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        fname : prepended name to plots fname_mname*
        mname : prepended name to plots fname_mname*
    """
    data, slabels = edata
    nlen = len(data)
    dlen = len(slabels)

    ddvar = data[0][:, 1:]
    ttime = data[0][:, 0].flatten()
    # tlen = len(ttime)
    for i in range(1, nlen):
        ddvar = np.concatenate((ddvar, data[i][:, 1:]), axis=0)
        ttime = np.concatenate((ttime, data[i][:, 0]))

    for j in range(dlen):
        ddd = ddvar[:, j].flatten()
        csq = int(np.sqrt(len(ddd)))
        zss, xss, yss = np.histogram2d(ttime, ddd, bins=(csq, csq))
        plt.figure(figsize=(9.68, 3.95))
        plt.xlabel("time")
        plt.ylabel("conc(" + slabels[j] + ")")
        delcon = int(np.max(zss) - np.min(zss)) / 500
        if slabels[j] == "P":
            plt.yscale('log')
        delcon = int(np.max(zss) - np.min(zss)) / 500
        try:
            cntr = plt.contour(xss[1:], yss[1:], zss.T,
                               list(range(int(np.min(zss)),
                                          int(np.max(zss)) + 1,
                                          int(delcon))
                                    ), cmap="Spectral_r")
        except:
            cntr = plt.contour(xss[1:], yss[1:], zss.T, 100, cmap="Spectral_r")
        fig = plt.gcf()
        fig.colorbar(cntr)
        plt.tight_layout()
        plt.savefig(
            fname +
            "_" +
            mname +
            "_" +
            slabels[j] +
            "_prob_density_wtime.jpg")
        plt.close()


def prob_density_calc_tslice(edata, bins=50, fname=""):
    """This function calculates the probability density of edata[0] per
    bins and returns a plot of the probability density.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        bins : number of bins an entire trajectory will be discretized
        fname : name prepended to plot name
    """
    if len(edata[0]) != 3:
        ndata, slabels = sample_points(edata)
        time = ndata[0]
        nlen = len(ndata)
        dlen = len(slabels)

        ddd = [[] for x in range(dlen)]
        for k in range(1, nlen):
            ddvar = ndata[k]
            for j in range(dlen):
                ddd[j].append(ddvar[j])
        ddd = np.array(ddd)

        vvv = {}
        for i in range(dlen):
            for j in range(len(time)):
                vvv[(i, j)] = np.histogram(ddd[i][:, j], bins=bins)

        tlen = len(time)
        strt = int(tlen / 10)
        for i in range(dlen):
            lines = []
            fig = plt.figure(figsize=(9.68, 3.95))
            axf = fig.gca(projection='3d')
            for j in range(strt, tlen, int(tlen / 5)):
                xss = vvv[(i, j)][1][1:]
                yss = vvv[(i, j)][0]
                width = (xss[1] - xss[0])
                line = axf.bar(xss, yss, zs=time[j], zdir='x',
                               width=max(width, 0.8), alpha=1)
                lines.append(line)
            axf.set_xlabel('time')
            axf.set_ylabel("conc(" + slabels[i] + ")")
            axf.set_zlabel('freq')
            axf.view_init(elev=40, azim=-120)
            plt.savefig(fname + "_" + slabels[j] + "_prob_density_tslice.jpg")
            plt.close()


def prob_density_calc(edata, fname):
    """This function calculates the probability density of edata[0] and
    returns a plot of the probability density.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        items : 3 item list of [canvas, scroll_x, scroll_y]
    """
    data, slabels = edata
    nlen = len(data)
    dlen = len(slabels)

    ddvar = data[0][:, 1:]
    for i in range(1, nlen):
        ddvar = np.concatenate((ddvar, data[i][:, 1:]), axis=0)

    for j in range(dlen):
        plt.figure(figsize=(9.68, 3.95))
        plt.xlabel("conc(" + slabels[j] + ")")
        plt.ylabel("Prob")
        ddd = ddvar[:, j]
        _ = plt.hist(ddd, bins=int(np.sqrt(len(ddd))),
                     density=True, orientation='vertical')
        plt.legend([slabels[j]])
        plt.xscale('log')
        plt.tight_layout()
        _ = plt.gcf()
        plt.savefig(fname + "_" + slabels[j] + "_prob_density.jpg")
        plt.close()



"""

                This is the param_est module

The following are the functions inside this module

1. load_data
2. custom_likelihood
3. ave_abs_dev
4. label_param
5. param_estimate


"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from tkinter import filedialog
import numpy as np
# from scipy.integrate import odeint
from scipy import optimize

from BioSANS2020.propagation.deterministic.ode_int import ode_int
from BioSANS2020.model.param_est.my_mcem import run_mcem
from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.gui_functs.scrollable_text import INSERT
# from BioSANS2020.myglobal import proc_global


def load_data(file=None):
    """This function reads the trajectory file containing the data which
    should follow the following example format;

    time	A	B
    0.0	100.0	0.0
    0.25	88.24969025197632	11.750309748023732
    0.5	77.88007831231087	22.119921687689185
    0.75	68.72892784164061	31.27107215835944
    1.0	60.65306592491437	39.346934075085684
    1.25	53.526142785532	46.473857214468055
    1.5	47.236655135816875	52.76334486418318
    1.75	41.68620193454698	58.31379806545308
    2.0	36.78794415253036	63.21205584746969
    2.25	32.46524678349081	67.53475321650924
    ...

    The format  above have  a header where the first  column is time and
    all other columns are species or components. The rows are the values
    of measurements  corresponding to  the header. Each row is delimited
    by  tab character "\t". If  the data file is in excel, just copy the
    data  from  excel to a text editor, save  it with a  filename and it
    will already  be tab delimited. If there  are several  replicates of
    the  data, just append them to  the end without  the header and this
    function can still handle all replicates.

    time	A	B
    0.0	100.0	0.0                                 # first replicate
    ...                                             # continuation
    0.0	100.0	0.0                                 # second replicate
    ...                                             # continuation

    Args:
        file (string, optional): trajectory file. Defaults to None.

    Returns:
        tuple: tuple  of  data and  labels (data, labels). The data is a
            list of all  the trajectories in the file and the labels are
            the corresponding name of the columns in the trajectory
    """
    end_time = 0
    if file is None:
        file = filedialog.askopenfilename(title="Select file")
    with open(file, "r") as fvar:
        data = []
        ddvar = []
        row1 = str(fvar.readline()).strip()
        slabels = row1.split("\t")[1:]
        for row in fvar:
            cols = [float(xvar) for xvar in row.split("\t")]

            # if time decreases, it is a signal of new trajectory.
            if end_time > cols[0] and ddvar:
                data.append(np.array(ddvar))
                ddvar = []
            ddvar.append(cols)
            end_time = cols[0]
        data.append(np.array(ddvar))
        return (data, slabels)


def custom_likelihood(ks_par, args=None):
    """This  function returns  the negative of the sum of squared errors
    between the true value of trajectory and estimated trajectory. This
    serves  as the fitness/objective/cost function  with a maximum value
    of zero.

    Args:
        ks_par (list): list of parameters
        args (tuple): Tuple  of  (data,  conc, tvar,  sp_comp,  ks_dict,
            r_dict,  p_dict,  v_stoich,  c_miss,  k_miss, molar, rfile).
            Defaults  to None. See param_estimate function for  the pro-
            per definition of each variables.

    Returns:
        float: negative of the sum of squared error
    """
    data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict, \
        v_stoich, c_miss, k_miss, molar, rfile = args
    ind = 0
    slabels, data = data
    for row in k_miss:
        i, j = row
        ks_dict[i][j] = ks_par[ind]
        ind = ind + 1
    for spi in c_miss:
        conc[spi] = ks_par[ind]
        ind = ind + 1
    # print(conc)
    z_val = ode_int(conc, tvar, sp_comp, ks_dict, r_dict,
                    p_dict, v_stoich, molar, rfile)
    spc1 = np.array(list(sp_comp.keys()))  # [spi for spi in sp_comp]
    # print(z_val[0,np.isin(spc1,c_miss)])
    spc1 = ~np.isin(spc1, c_miss)
    spc2 = np.array(slabels)
    spc2 = ~np.isin(spc2, c_miss)
    return -np.sum((data[:, spc2] - z_val[:, spc1])**2)


def ave_abs_dev(ks_par, args=None):
    """This  function  returns  the  mean of relative absolute deviation
    between the true value of trajectory and estimated trajectory. This
    serves  as the fitness/objective/cost function  with a maximum value
    of zero.

    Args:
        ks_par (list): list of parameters
        args (tuple): Tuple  of  (data,  conc, tvar,  sp_comp,  ks_dict,
            r_dict,  p_dict,  v_stoich,  c_miss,  k_miss, molar, rfile).
            Defaults  to None. See param_estimate function for  the pro-
            per definition of each variables.

    Returns:
        float: mean of relative absolute deviation
    """
    data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict, \
        v_stoich, c_miss, k_miss, molar, rfile = args
    ind = 0
    slabels, data = data
    for row in k_miss:
        i, j = row
        ks_dict[i][j] = ks_par[ind]
        ind = ind + 1
    for spi in c_miss:
        conc[spi] = ks_par[ind]
        ind = ind + 1
    z_val = ode_int(conc, tvar, sp_comp, ks_dict, r_dict,
                    p_dict, v_stoich, molar, rfile)
    spc1 = np.array(list(sp_comp.keys()))  # [spi for spi in sp_comp]
    spc1 = ~np.isin(spc1, c_miss)
    spc2 = np.array(slabels)
    spc2 = ~np.isin(spc2, c_miss)
    denom = data[:, spc2]
    denom[np.abs(denom) < 1.0e-10] = 1
    return np.mean(np.abs((data[:, spc2] - z_val[:, spc1]) / denom))


def label_param(k_miss, c_miss, ks_par):
    """[summary]

    Args:
        k_miss (list): list of position of unkwon parameters in the
            reaction tag of BioSANS topology file. For example;

            #REACTIONS
            A => B , -1      # forward rate constant is unknown
            B <=> C, -1, -1  # forward and backward is unknown

            The negative values means rate constant is unknown for those
            reaction.

            The value of k_miss will be;

                k_miss = [(0,0), (1, 0), (1, 1)]

            The format of the tuple is;

                (reaction index, rate constant index)
                (0, 0) - first reaction, first rate constant
                (1, 0) - second reaction, first rate constant
                (1, 1) - second reaction, second rate constant

        c_miss (list): component/species without trajectory data. They
            can be set in a topology file with negative values as well.

            @CONCENTRATION
            A, Ao
            B, -1
            C, Co

            The value of c_miss will be;

                c_miss = ['B']

        ks_par (list): list of estimated parameter values for k_miss as
            it appears sequencially and list of initial value for c_miss
            as it appears sequencially.

            Example;

                ks_par = [
                    k1f, # numeric value for position (0, 0)
                    k2f, # numeric value for position (1, 0)
                    k2b, # numeric value for position (1, 1)
                    Bo   # numeric initial value for B
                ]

    Returns:
        dictionary: dictionary  of estimated  values  with the following
        format {label : estimate,...}
    """
    pars = {}
    count = 0
    lvar = ["kf", "kb"]
    for xvar in k_miss:
        i, j = xvar
        pars[lvar[j] + str(i + 1)] = ks_par[count]
        count = count + 1

    for xvar in c_miss:
        pars[xvar + "o"] = ks_par[count]
        count = count + 1
    return pars


def param_estimate(conc, tvar, sp_comp, ks_dict, r_dict, p_dict,
                   v_stoich, items, molar=False, mode="MCEM",
                   true_data_fil=None, rfile=None):
    """[summary]

    Args:
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation

        tvar (list): time stamp of trajectories i.e. [0, 0.1, 0.2, ...]
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1

        ks_dict (dict): dictionary of rate constant that appears in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }

        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }

        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        v_stoich (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])

        items (list): list of [canvas, scroll_x, scroll_y]
        molar (bool, optional): If True, the units for any amount is in
            molar. Propensity will be macroscopic. Defaults to False.
        mode (str, optional): parameter estimation method keyword which
            is  one of "MCEM", "DEvol", "NeldMead", "Powell", "L-BFGS-B"
        true_data_fil (string, optional): file  name  of  the trajectory
            file containing experimental or observed trajectory.
            Defaults to None.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        tuple: (0, param_res) where param_res is a dictionary of estima-
        ted  values  with the following format {label : estimate,...}
    """

    data2 = load_data(true_data_fil)
    slabels = data2[1]
    spc = list(sp_comp.keys())  # [spi for spi in sp_comp]
    z_val = [conc[a] for a in sp_comp]
    c_miss = []
    for i, _ in enumerate(z_val):
        if z_val[i] < 0:
            c_miss.append(spc[i])
    k_miss = []
    for i, _ in enumerate(ks_dict):
        if len(ks_dict[i]) == 1:
            if ks_dict[i][0] < 0:
                k_miss.append((i, 0))
        elif len(ks_dict[i]) == 2:
            if ks_dict[i][0] < 0:
                k_miss.append((i, 0))
            if ks_dict[i][1] < 0:
                k_miss.append((i, 1))
    npar = len(c_miss) + len(k_miss)

    tvar = data2[0][0][:, 0]
    data = (slabels, data2[0][0][:, 1:])
    # print(data)

    if items:
        text = prepare_scroll_text(items)

        def ffprint(xvar):
            return text.insert(INSERT, " ".join([str(y) for y in xvar]))
    else:
        def ffprint(xvar):
            return print(" ".join([str(y) for y in xvar]))

    param_res = {}
    if mode == "MCEM":
        fvar = 1
        k = 10
        ks_par, er_min = run_mcem(
            fvar, npar, fvar, k * fvar, positive_only=True,
            likelihood=custom_likelihood,
            arg=(data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict,
                 v_stoich, c_miss, k_miss, molar, rfile)
        )

        error = ave_abs_dev(
            ks_par, (data, conc, tvar, sp_comp, ks_dict, r_dict,
                     p_dict, v_stoich, c_miss, k_miss, molar, rfile)
        )

        count = 0
        best_er = 1e+100
        best_ks = ks_par

        rands = [(xvar + 1) * np.random.uniform(0, 1)
                 for xvar in range(200000)]
        ind = 0
        while error > 1.0e-6 and k <= 10000:
            ks_par, er_min = run_mcem(
                min(fvar, 10), npar, fvar, k * fvar, positive_only=True,
                likelihood=custom_likelihood,
                arg=(data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict,
                     v_stoich, c_miss, k_miss, molar, rfile), r_rand=rands[ind]
            )

            error = ave_abs_dev(
                ks_par, (data, conc, tvar, sp_comp, ks_dict, r_dict,
                         p_dict, v_stoich, c_miss, k_miss, molar, rfile))
            if error < best_er:
                best_er = error
                best_ks = ks_par

                param_res = label_param(k_miss, c_miss, best_ks)
                ffprint(["\n\npartial result =\n", "errors = ", error, er_min])
                ffprint(["\n"])
                for xvar in param_res:
                    ffprint(["\n" + str(xvar) + " = " + str(param_res[xvar])])

            fvar = fvar + 1
            count = count + 1
            if count == 10:
                k = k * 10
                fvar = 1
                count = 0
            ind = ind + 1
        param_res = label_param(k_miss, c_miss, best_ks)
        ffprint(["\n\nFinal result =\n", best_ks, error, er_min])
        ffprint(["\n"])
        for xvar in param_res:
            ffprint(["\n" + str(xvar) + " = " + str(param_res[xvar])])
    elif mode == "DEvol":
        x_ini = []
        for _ in k_miss:
            x_ini.append(np.random.uniform())
        for _ in c_miss:
            x_ini.append(np.random.uniform())
        ddvar = optimize.minimize(
            lambda ks_par: -
            custom_likelihood(
                ks_par, (data, conc, tvar, sp_comp, ks_dict, r_dict,
                         p_dict, v_stoich, c_miss, k_miss, molar, rfile)
            ),
            x_ini, method='Nelder-Mead', tol=1e-10,
            options={
                'maxiter': 10000,
                'adaptive': True})

        bounds = []
        counts = 0
        for _ in k_miss:
            val = ddvar.x[counts]
            bounds.append((max(0, 0.1 * val), max(0, 10 * val)))
            counts = counts + 1
        for _ in c_miss:
            val = ddvar.x[counts]
            bounds.append((max(0, 0.1 * val), max(0, 10 * val)))
            counts = counts + 1

        ddvar = optimize.differential_evolution(
            lambda ks_par: -custom_likelihood
            (ks_par, (data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict,
                      v_stoich, c_miss, k_miss, molar, rfile)),
            bounds, maxiter=100, popsize=10, tol=1.0e-10)

        ddvar = optimize.minimize(
            lambda ks_par: -custom_likelihood
            (ks_par, (data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict,
                      v_stoich, c_miss, k_miss, molar, rfile)),
            ddvar.x, method='Nelder-Mead', tol=1e-10,
            options={'maxiter': 10000, 'adaptive': True})

        param_res = label_param(k_miss, c_miss, ddvar.x)
        ffprint(["\nFinal result =\n", ddvar])
        ffprint(["\n"])
        for xvar in param_res:
            ffprint(["\n" + str(xvar) + " = " + str(param_res[xvar])])
    elif mode == "NeldMead":
        x_ini = []
        for _ in k_miss:
            x_ini.append(np.random.uniform())
        for _ in c_miss:
            x_ini.append(np.random.uniform())

        ddvar = optimize.minimize(
            lambda ks_par: -custom_likelihood(
                ks_par, (data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict,
                         v_stoich, c_miss, k_miss, molar, rfile)), x_ini,
            method='Nelder-Mead', tol=1e-10,
            options={'maxiter': 100000, 'adaptive': True})

        param_res = label_param(k_miss, c_miss, ddvar.x)
        ffprint(["\nFinal result =\n", ddvar])
        ffprint(["\n"])
        for xvar in param_res:
            ffprint(["\n" + str(xvar) + " = " + str(param_res[xvar])])
    elif mode == "Powell":
        x_ini = []
        for _ in k_miss:
            x_ini.append(np.random.uniform())
        for _ in c_miss:
            x_ini.append(np.random.uniform())

        ddvar = optimize.minimize(
            lambda ks_par: -custom_likelihood(
                ks_par,
                (data, conc, tvar, sp_comp, ks_dict, r_dict,
                 p_dict, v_stoich, c_miss, k_miss, molar, rfile)
            ),
            x_ini, method='Powell', tol=1e-10,
            options={'maxiter': 100000, 'ftol': 1e-10, 'xtol': 1e-10})

        param_res = label_param(k_miss, c_miss, ddvar.x)
        ffprint(["\nFinal result =\n", ddvar])
        ffprint(["\n"])
        for xvar in param_res:
            ffprint(["\n" + str(xvar) + " = " + str(param_res[xvar])])
    elif mode == "L-BFGS-B":
        x_ini = []
        for _ in k_miss:
            x_ini.append(np.random.uniform())
        for _ in c_miss:
            x_ini.append(np.random.uniform())

        ddvar = optimize.minimize(
            lambda ks_par: -custom_likelihood
            (ks_par, (data, conc, tvar, sp_comp, ks_dict, r_dict, p_dict,
                      v_stoich, c_miss, k_miss, molar, rfile)
             ), x_ini, method='L-BFGS-B', tol=1e-10,

            options={'maxiter': 100000, 'ftol': 1e-10, 'xtol': 1e-10})

        param_res = label_param(k_miss, c_miss, ddvar.x)
        ffprint(["\nFinal result =\n", ddvar])
        ffprint(["\n"])
        for xvar in param_res:
            ffprint(["\n" + str(xvar) + " = " + str(param_res[xvar])])
    return (0, param_res)



"""

                  This module is the mytauleap module

The purpose of this module is to propagate stochastic trajectories using
the tau-leaping algorithm.

The functions in this module are the following;

1. sim_tauLeap
2.


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from BioSANS2020.propagation.propensity import propensity_vec
from BioSANS2020.myglobal import mglobals as globals2
from BioSANS2020.propagation.recalculate_globals \
    import get_globals, apply_rules, reserve_events_words


def sim_tauLeap(tvar, sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                rand_seed, del_coef, implicit=False, rfile=""):
    """This functions performs the tau-leaping integration

    Args:
        tvar ([type]): [description]
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
         p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        rand_seed ([type]): [description]
        rand_seed (float): random  seed value picked  at random for each
            trajectory. They have been sampled from the calling program.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        tuple : (time, trajectories)
    """
    get_globals(rfile)
    tmax = tvar[-1]
    np.random.seed(int(rand_seed * 100))
    tnew = []
    stch_var2 = stch_var**2
    stch_var = stch_var.T

    all_sp = list(sp_comp.keys())  # [z for z in sp_comp]
    spc = [z for z in sp_comp if z not in reserve_events_words]
    spc2 = [z for z in sp_comp if z in reserve_events_words]
    concz = {xvar: conc[xvar] for xvar in conc}
    yconc = {xvar: conc[xvar] for xvar in conc}
    apply_rules(concz, yconc)
    update_sp = [all_sp.index(z) for z in spc]
    zlist = [[concz[z] for z in spc]]
    t_c = 0
    tnew.append(t_c)

    g_i = []
    for i, _ in enumerate(spc):
        keys = sp_comp[spc[i]]
        maxrlen = 0
        rlen = 0
        for key in keys:
            if spc[i] in r_dict[key]:
                maxrlen = max(maxrlen, len(r_dict[key]))
            elif spc[i] in p_dict[key]:
                maxrlen = max(maxrlen, len(p_dict[key]))
        if maxrlen == 1:
            for key in keys:
                if spc[i] in r_dict[key]:
                    rlen = max(rlen, r_dict[key][spc[i]])
                elif spc[i] in p_dict[key]:
                    rlen = max(rlen, p_dict[key][spc[i]])
            if rlen == 1:
                g_i.append(lambda xvar: 1)
            else:
                g_i.append(lambda xvar: 2)
        elif maxrlen == 2:
            try:
                g_i.append(lambda xvar: 2 + 1 / (xvar - 1))
            except:
                g_i.append(lambda xvar: 2)
        else:
            g_i.append(lambda xvar: 1)

    if not implicit:
        while t_c < tmax:
            for xvar, _ in enumerate(spc):
                concz[spc[xvar]] = zlist[-1][xvar]
            prop_flux = propensity_vec(ks_dict, concz, r_dict, p_dict)
            alp = np.sum(prop_flux)
            r_1 = np.random.uniform()
            while r_1 == 0:
                r_1 = np.random.uniform()
            dt1 = (1 / alp) * (np.log(1 / r_1))

            uuj = np.matmul(stch_var, prop_flux)
            sig = np.matmul(stch_var2, prop_flux)
            exigi = np.array([zlist[-1][j] * (1 / g_i[j](zlist[-1][j]))
                              for j in range(len(spc))]) * 0.03 * del_coef
            exigi1 = np.maximum(exigi, np.full(len(spc), 1))
            dt2 = min(np.min(exigi1 / np.abs(uuj)),
                      np.min(exigi1 * exigi1 / np.abs(sig)))

            dtime = max(dt1, dt2)
            kvar = np.round(np.random.poisson(prop_flux * dtime))
            all_pos = True
            c_c = {}
            b_b = np.sum(kvar * stch_var[:, update_sp], 0)
            for xvar, _ in enumerate(spc):
                holder = zlist[-1][xvar] + b_b[xvar]
                if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                    c_c[spc[xvar]] = holder
                else:
                    all_pos = False
                    break
            if all_pos:
                for xvar, _ in enumerate(spc):
                    concz[spc[xvar]] = c_c[spc[xvar]]
                for xvar, _ in enumerate(spc2):
                    concz[spc2[xvar]] = concz[spc2[xvar]] + dtime
                apply_rules(concz, yconc)
                zlist.append([concz[xvar] for xvar in spc])
                t_c = t_c + dtime
                tnew.append(t_c)

    else:
        tindex = 1
        z_c = [zlist[-1]]
        while t_c < tmax:
            prop_flux = propensity_vec(ks_dict, concz, r_dict, p_dict)
            prop_flux[prop_flux < 0] = 0
            alp = np.sum(prop_flux)
            r_1 = np.random.uniform()
            while r_1 == 0:
                r_1 = np.random.uniform()
            dt1 = (1 / alp) * (np.log(1 / r_1))

            uuj = np.matmul(stch_var, prop_flux)
            sig = np.matmul(stch_var2, prop_flux)
            exigi = np.array([zlist[-1][j] * (1 / g_i[j](zlist[-1][j]))
                              for j in range(len(spc))]) * 0.03 * del_coef
            exigi1 = np.maximum(exigi, np.full(len(spc), 1))
            dt2 = min(np.min(exigi1 / np.abs(uuj)),
                      np.min(exigi1 * exigi1 / np.abs(sig)))

            dtime = max(dt1, dt2)
            if t_c + dtime > tvar[tindex]:
                dtime = max(0, tvar[tindex] - t_c)
                kvar = np.round(np.random.poisson(prop_flux * dtime))
                all_pos = True
                c_c = {}
                b_b = np.sum(kvar * stch_var[:, update_sp], 0)
                for xvar, _ in enumerate(spc):
                    holder = zlist[-1][xvar] + b_b[xvar]
                    if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                        c_c[spc[xvar]] = holder
                    else:
                        all_pos = False
                        break
                if all_pos:
                    for xvar, _ in enumerate(spc):
                        concz[spc[xvar]] = c_c[spc[xvar]]
                    for xvar, _ in enumerate(spc2):
                        concz[spc2[xvar]] = concz[spc2[xvar]] + dtime
                    apply_rules(concz, yconc)
                    zlist.append([concz[xvar] for xvar in spc])
                    z_c.append(zlist[-1])
                    t_c = t_c + dtime
                    tindex = tindex + 1
            else:
                dtime = max(0, dtime)
                kvar = np.round(np.random.poisson(prop_flux * dtime))
                all_pos = True
                c_c = {}
                b_b = np.sum(kvar * stch_var[:, update_sp], 0)
                for xvar, _ in enumerate(spc):
                    holder = zlist[-1][xvar] + b_b[xvar]
                    if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                        c_c[spc[xvar]] = holder
                    else:
                        all_pos = False
                        break
                if all_pos:
                    for xvar, _ in enumerate(spc):
                        concz[spc[xvar]] = c_c[spc[xvar]]
                    for xvar, _ in enumerate(spc2):
                        concz[spc2[xvar]] = concz[spc2[xvar]] + dtime
                    apply_rules(concz, yconc)
                    zlist.append([concz[xvar] for xvar in spc])
                    t_c = t_c + dtime

        if len(tvar) != len(z_c):
            while len(tvar) != len(z_c):
                z_c.append(zlist[-1])
        tnew = tvar
        zlist = z_c
    return (tnew, np.array(zlist))



"""

                  This module is the mystiffcle module

This can propagate non-stiff to moderately stiff stochastic simulation
using the chemical langevine equation. Here two versions are provided

1) Tau-adaptive CLE
2) Fix-inreval CLE

The following are the list of function for this module.

1. cle_model
2. cle_calculate
3. cle2_calculate


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from BioSANS2020.propagation.propensity import propensity_vec
from BioSANS2020.propagation.recalculate_globals \
    import get_globals, apply_rules, reserve_events_words
# from BioSANS2020.myglobal import mglobals as globals2


def cle_model(sp_comp, ks_dict, conc, r_dict, p_dict,
              stch_var, dtime, del_coef, reg=False):
    """This functions prepare the CLE model for integration

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        dtime (float): step-size
        del_coef (float): step-size factor or modifier
        reg (bool, optional): If True, the model is for fix-interval CLE
            . Defaults to False.

    Returns:
        np.ndarray: f_d = fofx * dtime + gofx * sqdt
    """
    prop_flux = propensity_vec(ks_dict, conc, r_dict, p_dict)
    sqrt_prop = np.sqrt(prop_flux)
    nlen = np.random.randn(len(prop_flux)).reshape(len(prop_flux), 1)
    fofx = np.matmul(stch_var, prop_flux)
    gofx = np.matmul(stch_var, sqrt_prop * nlen)
    if not reg:
        h_1 = np.abs(prop_flux) + 1.0e-30  # np.abs(fofx)+1.0e-30
        h_2 = np.abs(sqrt_prop) + 1.0e-30  # np.abs(gofx)+1.0e-30
        dtime = min(del_coef * min(np.min(1 / h_1), np.min(1 / h_2)), dtime)
    sqdt = np.sqrt(dtime)

    ind = 0
    for s_p in sp_comp:
        key = s_p.strip().split("_")[0]
        if key in reserve_events_words:
            gofx[ind] = 0
        ind = ind + 1

    f_d = fofx * dtime + gofx * sqdt
    return [f_d.reshape(len(sp_comp)), dtime]


def cle_calculate(tvar, sp_comp, ks_dict, sconc, r_dict, p_dict, stch_var,
                  del_coef=10, rand_seed=1, implicit=False, rfile=""):
    """This functions performs the tau-adaptive CLE integration
    Args:
        tvar (list): time stamp of simulation
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        sconc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        del_coef (float): step-size factor or modifier
        rand_seed (float): random  seed value picked  at random for each
            trajectory. They have been sampled from the calling program.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        tuple: (time, trajectories)
    """
    get_globals(rfile)
    np.random.seed(int(rand_seed * 100))
    tnew = []
    dtime = tvar[-1] - tvar[-2]
    yconc = {x: sconc[x] for x in sconc}
    conc = {x: sconc[x] for x in sconc}
    apply_rules(conc, yconc)
    res_traj = [[conc[z] for z in sp_comp]]

    if not implicit:
        tnow = tvar[0]
        tnew.append(tnow)
        while tnow < tvar[-1]:
            mvar = np.nan_to_num(
                cle_model(sp_comp, ks_dict, conc, r_dict, p_dict,
                          stch_var, dtime, del_coef))
            mmvar = mvar[0].reshape(1, len(mvar[0]))[0]
            tnow = tnow + mvar[1]
            ind = 0
            for s_p in sp_comp:
                conc[s_p] = res_traj[-1][ind] + mmvar[ind]
                ind = ind + 1
            apply_rules(conc, yconc)
            res_traj.append([conc[z] for z in sp_comp])
            for s_p in sp_comp:
                conc[s_p] = max(0, conc[s_p])
            tnew.append(tnow)
    else:
        tnow = tvar[0]
        tnew.append(tnow)
        tindex = 1
        cvar = [conc[z] for z in sp_comp]
        while tnew[-1] < tvar[-1]:
            mvar = np.nan_to_num(
                cle_model(sp_comp, ks_dict, conc, r_dict, p_dict,
                          stch_var, dtime, del_coef))
            tnow = tnow + mvar[1]
            if tnow > tvar[tindex]:
                tnow = tnow - mvar[1]
                dt2 = tvar[tindex] - tnow
                mvar = np.nan_to_num(
                    cle_model(sp_comp, ks_dict, conc, r_dict, p_dict,
                              stch_var, dt2, del_coef, True))
                mmvar = mvar[0].reshape(1, len(mvar[0]))[0]
                tnow = tnow + mvar[1]

                ind = 0
                for s_p in sp_comp:
                    conc[s_p] = cvar[ind] + mmvar[ind]
                    ind = ind + 1
                apply_rules(conc, yconc)
                cvar = [conc[z] for z in sp_comp]
                for s_p in sp_comp:
                    conc[s_p] = max(0, conc[s_p])
                res_traj.append(cvar)

                tnew.append(tnow)
                tindex = tindex + 1
            else:
                mmvar = mvar[0].reshape(1, len(mvar[0]))[0]
                ind = 0
                for s_p in sp_comp:
                    conc[s_p] = cvar[ind] + mmvar[ind]
                    ind = ind + 1
                apply_rules(conc, yconc)
                cvar = [conc[z] for z in sp_comp]
                for s_p in sp_comp:
                    conc[s_p] = max(0, conc[s_p])
    return (tnew, np.array(res_traj))


def cle2_calculate(tvar, sp_comp, ks_dict, sconc, r_dict, p_dict,
                   stch_var, del_coef=1, rand_seed=1, rfile=""):
    """This functions performs the fix-interval CLE integration
    Args:
        tvar (list): time stamp of simulation
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        sconc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        del_coef (float): step-size factor or modifier
        rand_seed (float): random  seed value picked  at random for each
            trajectory. They have been sampled from the calling program.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        tuple: (time, trajectories)
    """
    get_globals(rfile)
    np.random.seed(int(rand_seed * 100))
    div = max(1, int(1 / del_coef))
    dtime = (tvar[-1] - tvar[-2]) / div
    yconc = {x: sconc[x] for x in sconc}
    conc = {x: sconc[x] for x in sconc}
    apply_rules(conc, yconc)
    res_traj = [[conc[z] for z in sp_comp]]
    tnow = tvar[0]
    tnew = [tnow]
    d_v = 0
    cvar = res_traj[-1]
    while abs(tnow - tvar[-1]) > 1.0e-10:
        mvar = np.nan_to_num(
            cle_model(sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                      dtime, 1, True))
        mmvar = mvar[0].reshape(1, len(mvar[0]))[0]
        tnow = tnow + mvar[1]
        ind = 0
        for s_p in sp_comp:
            conc[s_p] = cvar[ind] + mmvar[ind]
            ind = ind + 1
        apply_rules(conc, yconc)
        cvar = [conc[z] for z in sp_comp]
        if d_v == div - 1:
            res_traj.append(cvar)
            tnew.append(tnow)
            d_v = 0
        else:
            d_v = d_v + 1
        for s_p in sp_comp:
            conc[s_p] = max(0, conc[s_p])

    return (tnew, np.array(res_traj))



"""

                    This module is the new_file module

    This module helps in displaying a new text area for typing topology,
codes, or regular text.

"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text


def new_file(items):
    """This function returns a blank text area for typing
    Args:
        items (tuple): (canvas, scroll_x, scroll_y)

    Returns:
        tkinter.Text: text area where a user can type
    """
    text = prepare_scroll_text(items)
    return text



"""

                   This module is the mglobals module

Variables global to  all modules are listed here and instantiated in one
of the main calling module such as BioSANS module and BioSSL module.

"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

PLOT_I = 0
INT_VARS = []
CONTAINER = []
PLOTTED = []
MODIFIED = {}
PROP_MODIFIED = {}
CON_BOUNDARY = {}
TO_CONVERT = ""
SETTINGS = {}
EXEC_FUNCTIONS = []
DELAY_LIST = {}
TCHECK = []


def init(self):
    """Re initiate global variable just in case needed
    """
    self.PLOT_I = 0
    self.INT_VARS = []
    self.CONTAINER = []
    self.PLOTTED = []
    self.MODIFIED = {}
    self.PROP_MODIFIED = {}
    self.CON_BOUNDARY = {}
    self.TO_CONVERT = ""
    self.SETTINGS = {}
    self.EXEC_FUNCTIONS = []
    self.DELAY_LIST = {}
    self.TCHECK = []



# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

"""
This module interpoltaes nsamp points from the trajectory.
"""

from numpy import linspace, interp


def sample_points(edata, nsamp=50000):
    """This function takes edata which is composed of numerical data and
    their labels and returns interpolated points.
    Args:
        edata : two dimensional array of data & labels (data, label).
                data is a 3D array where each row are the individual
                trajectories. Each trajectory is a 2D numpy array where
                the first column is time and the remaining columns are
                the corresponding components.
        nsamp : number of points to sample from data
    Returns:
        sdata : interpolated data points with sdata[0] is time and the
                remaining rows are interpolated data that corresponds to
                the time in sdata[0]
        label_name : labels
    """

    data, label_name = edata
    end_time = round(data[0][:, 0][-1])
    time = linspace(0, end_time, nsamp)
    sdata = [time]
    for ddrow in data:
        drow = []
        n_count = ddrow.shape[1]
        for i_index in range(1, n_count):
            y_interpolated = interp(time, ddrow[:, 0], ddrow[:, i_index])
            drow.append(y_interpolated)
        sdata.append(drow)
    return (sdata, label_name)



"""

                 This module is the proc_global module

The  main  purpose of this module is  to observed module wide processes.

"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import multiprocessing as mp


MANAGER = None
LST = None


def init(self):
    """Initiate process related global variables
    """
    self.MANAGER = mp.Manager()
    self.LST = self.MANAGER.list()



"""

				This is the topology_view module

	This module helps to load topology files into a text area for
displaying in BioSANS.

"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.gui_functs.scrollable_text import INSERT


def view_topo(topo, items):
    """This function open topology files and display the topology into a
    text area for further simulation or for editing.

    Args:
        topo (string): topology file name
        items (tuple): (canvas, scroll_x, scroll_y)

    Returns:
        tkinter.Text: text area where topology is displayed. A user can
            type and modify the topology
    """
    text = prepare_scroll_text(items)

    def ffprint(xvar):
        return text.insert(INSERT, " ".join([str(y) for y in xvar]))
    with open(topo) as ffile:
        for xvar in ffile:
            ffprint([xvar.replace("\t", " " * 4)])
    return text



"""

                  This module in the sbml_math module

This contains function needed to interpret SBML files.

"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import sympy
import numpy as np

# def abs(xvar):
# return sympy.fabs(xvar)
# need to use trigonometric relationship to create non-built in numpy
# trigonometric functions

NUMBER_TYPE = (float, int)


def acos(xvar):
    """returns sympy.acos(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arccos(xvar)
    return sympy.acos(xvar).evalf()


def arccos(xvar):
    """returns sympy.acos(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arccos(xvar)
    return sympy.acos(xvar).evalf()


def acosh(xvar):
    """returns sympy.acosh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arccosh(xvar)
    return sympy.acosh(xvar).evalf()


def arccosh(xvar):
    """returns sympy.acosh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arccosh(xvar)
    return sympy.acosh(xvar).evalf()


def acot(xvar):
    """returns sympy.acot(xvar).evalf()"""
    return sympy.acot(xvar).evalf()


def arccot(xvar):
    """returns sympy.acot(xvar).evalf()"""
    return sympy.acot(xvar).evalf()


def acoth(xvar):
    """returns sympy.acoth(xvar).evalf()"""
    return sympy.acoth(xvar).evalf()


def arccoth(xvar):
    """returns sympy.acoth(xvar).evalf()"""
    return sympy.acoth(xvar).evalf()


def acsc(xvar):
    """returns sympy.acsc(xvar).evalf()"""
    return sympy.acsc(xvar).evalf()


def arccsc(xvar):
    """returns sympy.acsc(xvar).evalf()"""
    return sympy.acsc(xvar).evalf()


def acsch(xvar):
    """returns sympy.acsch(xvar).evalf()"""
    return sympy.acsch(xvar).evalf()


def arccsch(xvar):
    """returns sympy.acsch(xvar).evalf()"""
    return sympy.acsch(xvar).evalf()


def arcsec(xvar):
    """returns sympy.asec(xvar).evalf()"""
    return sympy.asec(xvar).evalf()


def asech(xvar):
    """returns sympy.asech(xvar).evalf()"""
    return sympy.asech(xvar).evalf()


def arcsech(xvar):
    """returns sympy.asech(xvar).evalf()"""
    return sympy.asech(xvar).evalf()


def asin(xvar):
    """returns sympy.asin(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arcsin(xvar)
    return sympy.asin(xvar).evalf()


def asinh(xvar):
    """returns sympy.asinh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arcsinh(xvar)
    return sympy.asinh(xvar).evalf()


def arcsinh(xvar):
    """returns sympy.asinh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arcsinh(xvar)
    return sympy.asinh(xvar).evalf()


def arcsin(xvar):
    """returns sympy.asin(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arcsin(xvar)
    return sympy.asin(xvar).evalf()


def atan(xvar):
    """returns sympy.atan(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arctan(xvar)
    return sympy.atan(xvar).evalf()


def arctan(xvar):
    """returns sympy.atan(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arctan(xvar)
    return sympy.atan(xvar).evalf()


def atanh(xvar):
    """returns sympy.atanh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arctanh(xvar)
    return sympy.atanh(xvar).evalf()


def arctanh(xvar):
    """returns sympy.atanh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.arctanh(xvar)
    return sympy.atanh(xvar).evalf()


def ceil(xvar):
    """returns sympy.ceiling(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.ceil(xvar)
    return sympy.ceiling(xvar).evalf()


def ceiling(xvar):
    """returns sympy.ceiling(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.ceil(xvar)
    return sympy.ceiling(xvar).evalf()


def cos(xvar):
    """returns sympy.cos(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.cos(xvar)
    return sympy.cos(xvar).evalf()


def cosh(xvar):
    """returns sympy.cosh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.cosh(xvar)
    return sympy.cosh(xvar).evalf()


def cot(xvar):
    """returns sympy.cot(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return 1 / np.tan(xvar)
    return sympy.cot(xvar).evalf()


def coth(xvar):
    """returns sympy.coth(xvar).evalf()"""
    return sympy.coth(xvar).evalf()


def csc(xvar):
    """returns sympy.csc(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return 1 / np.sin(xvar)
    return sympy.csc(xvar).evalf()


def csch(xvar):
    """returns sympy.csch(xvar).evalf()"""
    return sympy.csch(xvar).evalf()


def factorial(xvar):
    """returns sympy.factorial(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.math.factorial(xvar)
    return sympy.factorial(xvar).evalf()


def exp(xvar):
    """returns sympy.exp(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.exp(xvar)
    return sympy.exp(xvar).evalf()


def floor(xvar):
    """returns sympy.floor(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.floor(xvar)
    return sympy.floor(xvar).evalf()


def ln(xvar):
    """returns sympy.ln(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.log(xvar)
    return sympy.ln(xvar).evalf()


def log(xvar):
    """returns sympy.log(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.log(xvar)
    return sympy.log(xvar).evalf()


def log10(xvar):
    """returns sympy.log(xvar, 10).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.log10(xvar)
    return sympy.log(xvar, 10).evalf()


def piecewise(*xvar):
    """returns the value before the first True value in xvar. If there
    is no True value, returns the last element in vxar"""
    if len(xvar) == 3:
        return xvar[0] if xvar[1] else xvar[2]

    # ans = []
    xlen = len(xvar) - 1
    for i in range(0, xlen, 2):
        if xvar[i + 1]:
            return xvar[i]
    return xvar[-1]


def pow(xvar, yvar):
    """sympy.Pow(xvar, yvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE) and isinstance(yvar, NUMBER_TYPE):
        return np.float_power(xvar, yvar)
    return sympy.Pow(xvar, yvar).evalf()


def power(xvar, yvar):
    """sympy.Pow(xvar, yvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE) and isinstance(yvar, NUMBER_TYPE):
        return np.float_power(xvar, yvar)
    return sympy.Pow(xvar, yvar).evalf()


def root(nvar, xvar):
    """sympy.root(xvar, nvar).evalf()"""
    return sympy.root(xvar, nvar).evalf()


def sec(xvar):
    """returns sympy.sec(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return 1 / np.cos(xvar)
    return sympy.sec(xvar).evalf()


def sech(xvar):
    """sympy.sech(xvar).evalf()"""
    return sympy.sech(xvar).evalf()


def sqr(xvar):
    """sympy.sqrt(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.sqrt(xvar)
    return sympy.sqrt(xvar).evalf()


def sqrt(xvar):
    """sympy.sqrt(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.sqrt(xvar)
    return sympy.sqrt(xvar).evalf()


def sin(xvar):
    """sympy.sin(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.sin(xvar)
    return sympy.sin(xvar).evalf()


def sinh(xvar):
    """sympy.sinh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.sinh(xvar)
    return sympy.sinh(xvar).evalf()


def tan(xvar):
    """sympy.tan(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.tan(xvar)
    return sympy.tan(xvar).evalf()


def tanh(xvar):
    """sympy.tanh(xvar).evalf()"""
    if isinstance(xvar, NUMBER_TYPE):
        return np.tanh(xvar)
    return sympy.tanh(xvar).evalf()


def And(*xvar):
    """returns True if all elements of xvar is True else returns False"""
    for yvar in xvar:
        # if yvar == False:
        if not yvar:
            return False
    return True


def Not(xvar):
    """returns the reverse of the boolean value of xvar"""
    return not xvar


def Or(*xvar):
    """returns True if at least one value in xvar is True else False"""
    for yvar in xvar:
        # if yvar == True:
        if yvar:
            return True
    return False


def xor(*xvar):
    """returns True if there is odd number of True else returns False"""
    odd = 0
    for yvar in xvar:
        # if yvar == True:
        if yvar:
            odd = odd + 1
    if odd % 2 == 0:
        return False

    return True


def eq(*xvar):
    """returns True if xvar[0] == xvar[1:] else returns False"""
    last = None
    for yvar in xvar:
        if last is None:
            last = yvar
        else:
            if yvar != last:
                return False
    return True


def geq(*xvar):
    """returns True if xvar[0] >= xvar[1:] else returns False"""
    last = None
    for yvar in xvar:
        if last is None:
            last = yvar
        else:
            if last < yvar:
                return False
    return True


def gt(*xvar):
    """returns True if xvar[0] > xvar[1:] else returns False"""
    last = None
    for yvar in xvar:
        if last is None:
            last = yvar
        else:
            if last <= yvar:
                return False
    return True


def leq(*xvar):
    """returns True if xvar[0] <= xvar[1:] else returns False"""
    last = None
    for yvar in xvar:
        if last is None:
            last = yvar
        else:
            if last > yvar:
                return False
    return True


def lt(*xvar):
    """returns True if xvar[0] < xvar[1:] else returns False"""
    last = None
    for yvar in xvar:
        if last is None:
            last = yvar
        else:
            if last >= yvar:
                return False
    return True


def neq(xvar, yvar):
    """returns True if xvar != yvar else returns False"""
    if xvar == yvar:
        return False
    return True


def plus(*xvar):
    """sum(xvar).evalf()"""
    return sum(xvar).evalf()


def times(*xvar):
    """returns the product of all elements in the list xvar"""
    pvar = 1
    for yvar in xvar:
        pvar = pvar * yvar
    return pvar


def minus(xvar, yvar):
    """returns xvar - yvar"""
    return xvar - yvar


def divide(xvar, yvar):
    """returns xvar / yvar"""
    return xvar / yvar


def multiply(*xvar):
    """returns the product of all elements in the list xvar"""
    return times(*xvar)


SBML_FUNCT_DICT = {
    "acos": acos, "arccos": arccos, "acosh": acosh, "arccosh": arccosh,
    "acot": acot, "arccot": arccot, "acoth": acoth, "arccoth": arccoth,
    "acsc": acsc, "arccsc": arccsc, "acsch": acsch, "arccsch": arccsch,
    "arcsec": arcsec, "asech": asech, "arcsech": arcsech, "asin": asin,
    "asinh": asinh, "arcsinh": arcsinh, "arcsin": arcsin, "atan": atan,
    "arctan": arctan, "atanh": atanh, "arctanh": arctanh, "ceil": ceil,
    "ceiling": ceiling, "cos": cos, "cosh": cosh, "cot": cot,
    "coth": coth, "csc": csc, "csch": csch, "factorial": factorial,
    "exp": exp, "floor": floor, "ln": ln, "log": log, "log10": log10,
    "piecewise": piecewise, "pow": pow, "power": power, "root": root,
    "sec": sec, "sech": sech, "sqr": sqr, "sqrt": sqrt, "sin": sin,
    "sinh": sinh, "tan": tan, "tanh": tanh, "And": And, "Not": Not,
    "Or": Or, "xor": xor, "eq": eq, "geq": geq, "gt": gt, "leq": leq,
    "lt": lt, "neq": neq, "plus": plus, "times": times, "minus": minus,
    "divide": divide, "multiply": multiply, "exponentiale": exp(1),
    "INF": np.inf, "NAN": np.nan, "inf": np.inf, "nan": np.nan
}



"""

                This module is the analytical_sol module

This module handles analytical expression derivation

The following are the functions in this module

1. solve_with_timeout
2. get_sets
3. grab_steady_state
4. cs_to_csr
5. analyt_soln

"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from func_timeout import func_timeout, FunctionTimedOut
from sympy import simplify, Matrix, solve, Add, integrate, Function, \
    dsolve, Symbol, exp, Eq, zeros
# import time

from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.propagation.propensity import propensity_vec_molar
# from BioSANS2020.myglobal import mglobals as globals2
from BioSANS2020.propagation.recalculate_globals import get_globals


def solve_with_timeout(f_e, j_s):
    """resturn solution to the system of equation in f_e with variables
    symbos in the list j_s

    Args:
        f_e (list): equations
        j_s (list): Symbols

    Returns:
        dict: sometimes list but normally {variable : value}
    """
    sol = None
    sol = solve(f_e, j_s)
    return sol


def get_sets(r_dict, p_dict):
    """[summary]

    Args:
        r_dict (dict): dictionary of reactants
        p_dict (dict): dictionary of products

    Returns:
        set: sets to sum that will be used as constraints
    """
    tosum = set()
    sets = [tosum]
    used = set()
    for ih_var, _ in enumerate(r_dict):
        used2 = set()
        if len(r_dict[ih_var]) == 1:
            for svar in sets:
                key = list(r_dict[ih_var].keys())[0]
                if key not in used and r_dict[ih_var][key] != 0:
                    svar.add(key)
                    used2.add(key)
        else:
            for j, _ in enumerate(sets):
                svar = sets[j]
                sets.append(svar.copy())
                key = list(r_dict[ih_var].keys())[0]
                if key not in used and r_dict[ih_var][key] != 0:
                    svar.add(key)
                    used2.add(key)
                key = list(r_dict[ih_var].keys())[1]
                if key not in used and r_dict[ih_var][key] != 0:
                    sets[-1].add(key)
                    used2.add(key)

        if len(p_dict[ih_var]) == 1:
            for svar in sets:
                key = list(p_dict[ih_var].keys())[0]
                if key not in used and p_dict[ih_var][key] != 0:
                    svar.add(key)
                    used2.add(key)
        else:
            for j, _ in enumerate(sets):
                svar = sets[j]
                sets.append(svar.copy())
                key = list(p_dict[ih_var].keys())[0]
                if key not in used and p_dict[ih_var][key] != 0:
                    svar.add(key)
                    used2.add(key)
                key = list(p_dict[ih_var].keys())[1]
                if key not in used and p_dict[ih_var][key] != 0:
                    sets[-1].add(key)
                    used2.add(key)
        for z_s in used2:
            used.add(z_s)
    return sets


def grab_steady_state(r_dict, p_dict, c_s, c_so, da_dt, j_s):
    """This function return steady state amounts of components

    Args:
        r_dict (dict): reactants dictionary
        p_dict (dict): products dictionary
        c_s (dict): dictionary of concentration Symbols/values
        c_so (dict): dictionary of initial concentration Symbols/values
        da_dt ([type]): ode expression
        j_s ([type]): list of components

    Returns:
        Matrix: steady state result
    """
    sets = get_sets(r_dict, p_dict)
    fez = [da_dt]
    for svar in sets:
        fez.append(sum([c_s[xvar] - c_so[xvar] for xvar in svar]))
    val2 = solve(fez, list(j_s))
    x_s = Matrix([val2[j_s[ih_var]] for ih_var in range(len(j_s))])
    return x_s


def cs_to_csr(f_x, c_s, cs_r, not_semi):
    """[summary]

    Args:
        f_x (expression): sympy expression
        c_s (Symbol): sympy symbols
        cs_r (Symbol): equivalent of c_s
        not_semi (boolean): if True, use pure symbolic

    Returns:
        expression: substituted expression
    """
    if not_semi:
        r_r = []
        for xvar in c_s:
            r_r.append((c_s[xvar], cs_r[xvar]))
        return f_x.subs(r_r)
    return f_x


def analyt_soln(sp_comp, ks_dict, conc, r_dict, p_dict, stch_var, items=None,
                rfile="", not_semi=True, mode=None):
    """This function finds an analytical solution to a system of ODE.

    Args:
        sp_comp (dict): dictionary of components
        ks_dict (dict):  dictionary of rate constant
        conc (dict): dictionary of concentrations
        r_dict (dict): dictionary of reactants
        p_dict (dict): dictionary of products
        stch_var (Matrix): stoichiomentric matrix
        items (tuple, optional): (canvas, scroll_x, scrolly).
        rfile (str, optional): BioSANS topology file.
        not_semi (bool, optional): if True, use pure symbolic solution;
                which is a function of time, initial concentration, and
                rate constant. Defaults to True.
        mode (str, optional): mode of computation;
                ftxo - function of time and initial condition
                ftks - function of time and rate constant
            Defaults to None.

    Returns:
        dict: dictionary of solution
    """
    get_globals(rfile)
    my_equations = {}

    if items:
        text = prepare_scroll_text(items)

        def ffprint(xvar): return text.insert(
            'insert', " ".join([str(yvar) for yvar in xvar]))
    elif items == 0:
        def ffprint(xvar):
            return print("", end="")
    else:
        def ffprint(xvar):
            return print(" ".join([str(yvar) for yvar in xvar]), end="")

    if not_semi:
        ffprint(["\n\nComplex Analytical expressions\n"])
        ffprint(
            ["\n\nThe complex expression is because sympy do not know how",
             "you want to simplify the expression\n\n"])
    else:
        ffprint(["Simple semi-analytical expression\n\n"])

    c_s = {}
    c_so = {}
    tvar = Symbol('t', negative=False, real=True)

    equivals = [(tvar, 100)]
    cs_r = {}
    for xvar in sp_comp:
        c_s[xvar] = Function(xvar, negative=False, real=True)
        c_s[xvar] = c_s[xvar](tvar)
        cs_r[xvar] = Symbol(xvar, negative=False, real=True)
        if mode == "ftxo":
            c_so[xvar] = Symbol(xvar + "o", negative=False, real=True)
        elif mode == "ftks":
            c_so[xvar] = conc[xvar]
        else:
            c_so[xvar] = Symbol(xvar + "o", negative=False,
                                real=True) if not_semi else conc[xvar]
        equivals.append((c_so[xvar], conc[xvar]))

    kcs = []
    for i, _ in enumerate(ks_dict):
        row = []
        if len(ks_dict[i]) == 1:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, negative=False, real=True))
            equivals.append((row[0], ks_dict[i][0]))
        else:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, negative=False, real=True))
            equivals.append((row[0], ks_dict[i][0]))
            key = 'kb' + str(i + 1)
            row.append(Symbol(key, negative=False, real=True))
            equivals.append((row[1], ks_dict[i][1]))
        kcs.append(row)

    if mode == "ftks":
        fvar = Matrix(propensity_vec_molar(kcs, c_s, r_dict, p_dict, True))
    elif mode == "ftxo":
        fvar = Matrix(propensity_vec_molar(ks_dict, c_s, r_dict, p_dict, True))
    else:
        fvar = Matrix(propensity_vec_molar(kcs, c_s, r_dict, p_dict, True)) \
            if not_semi else Matrix(
                propensity_vec_molar(ks_dict, c_s, r_dict, p_dict, True))

    stch_var = Matrix(stch_var)
    # c_s might have change after call
    for xvar in sp_comp:
        c_s[xvar] = Function(xvar, negative=False, real=True)
        c_s[xvar] = c_s[xvar](tvar)
    slabels = list(c_s.keys())  # [xvar for xvar in c_s]

    def stoich(xvar):
        return max(1, abs(stch_var[slabels.index(xvar)][0]))

    ss_var = []
    nz_var = []
    for row in range(stch_var.shape[0]):
        if sum(abs(stch_var[row, :])) != 0 and slabels[row][0] != "-":
            ss_var.append(list(stch_var[row, :]))
            nz_var.append(row)
    ss_var = Matrix(ss_var)

    da_dt = ss_var * fvar
    ccs = [c_s[xvar] for xvar in c_s]
    ccso = [c_so[xvar] for xvar in c_so]
    j_s = [ccs[xvar] for xvar in nz_var]
    jso = [ccso[xvar] for xvar in nz_var]

    da_dt2 = []
    for xvar in j_s:
        da_dt2.append(xvar.diff(tvar))

    fes = [Eq(da_dt2[i], da_dt[i]) for i in range(len(da_dt2))]

    left_hand = da_dt.jacobian(j_s)
    # print(left_hand)
    right_hand = da_dt - left_hand * Matrix(j_s)
    # print(right_hand)
    # print(fes)
    # print(da_dt)
    # print(j_s)

    check = True
    work = False
    for i in range(left_hand.shape[0]):
        if sum(abs(left_hand.col(i))) == 0:
            check = False

    if check:
        print("Inside Check")
        left_hand_j = np.array(left_hand)
        # print(left_hand_j)
        left_hand_j = Matrix(left_hand_j.flatten()).jacobian(j_s)
        # print(left_hand_j)
        if left_hand_j != zeros(left_hand_j.shape[0], left_hand_j.shape[1]):
            sets = get_sets(r_dict, p_dict)
            fe2 = []
            for svar in sets:
                fe2.append(
                    Eq(sum([c_s[xvar] / stoich(xvar) - c_so[xvar]
                            / stoich(xvar) for xvar in svar]), 0))
            xret = solve(fe2, j_s)
            # print(xret)
            # print(fes)

            for svar in xret:
                # print(fes[-1].atoms(Function))
                fes[-1] = Eq(fes[-1].lhs, fes[-1].rhs.subs(svar, xret[svar]))
            denom = cs_to_csr(fes[-1].rhs, c_s, cs_r, not_semi)
            itvar = cs_to_csr(j_s[-1], c_s, cs_r, not_semi)
            # pot_an = integrate(1/denom,(itvar,jso[-1],j_s[-1]))-tvar
            pot_an = integrate(
                1 / fes[-1].rhs, (j_s[-1], jso[-1], j_s[-1])) - tvar
            fe2.append(pot_an)
            try:
                sol = func_timeout(20, solve, args=(fe2, j_s))
                for c_c in sol:
                    rval = Matrix(c_c).subs(equivals)
                    if np.sum(np.array(rval) < 0) > 0:
                        pass
                    else:
                        x_sol = [c_c[xvar] for xvar in range(len(j_s))]
                        ind = 0
                        for xvar in x_sol:
                            var_sp = str(j_s[ind])
                            answer = str(simplify(xvar))
                            ffprint([var_sp, " = ", answer, "\n\n"])
                            ind = ind + 1
                            if var_sp not in my_equations:
                                my_equations[var_sp] = answer
                        break
                work = True
            except FunctionTimedOut:
                work = False
        else:
            print("Attempting Matrix method")
            svar = Symbol('s', negative=False, real=True)
            x_o = Matrix(jso)
            if sum(abs(right_hand)) == 0:
                # print("llll",1)
                try:
                    x_s = grab_steady_state(
                        r_dict, p_dict, c_s, c_so, da_dt, j_s)
                    try:
                        exp_d = func_timeout(
                            200, exp, args=(left_hand * tvar,))
                        x_sol = x_s + exp_d * (x_o - x_s)
                        ind = 0
                        for xvar in x_sol:
                            var_sp = str(j_s[ind])
                            answer = str(simplify(xvar))
                            ffprint([var_sp, " = ", answer, "\n\n"])
                            ind = ind + 1
                            if var_sp not in my_equations:
                                my_equations[var_sp] = answer
                        work = True
                    except FunctionTimedOut:
                        work = False
                except:
                    pass
            else:
                # print("llll",2)
                try:
                    try:
                        x_s = simplify(-left_hand**-1 * right_hand)
                        x_sol = x_s + exp(left_hand * tvar) * (x_o - x_s)
                    except:
                        x_sol = simplify(
                            exp(left_hand * tvar) * x_o
                            + integrate(exp(left_hand * tvar)
                                        * right_hand, (svar, 0, tvar)))
                    ind = 0
                    for xvar in x_sol:
                        var_sp = str(j_s[ind])
                        answer = str(simplify(xvar))
                        ffprint([var_sp, " = ", answer, "\n\n"])
                        ind = ind + 1
                        if var_sp not in my_equations:
                            my_equations[var_sp] = answer
                    work = True
                except:
                    pass

    if not work:
        used3 = {}
        try:
            print(1)
            try:
                # sol = dsolve(fes, j_s, ics={j_s[j].subs(tvar,0):jso[j]
                #                             for j in range(len(j_s))})
                sol = func_timeout(
                    200, lambda: dsolve(
                        fes, j_s, ics={j_s[j].subs(tvar, 0): jso[j]
                                       for j in range(len(j_s))}))
                for xvar, _ in enumerate(sol):
                    var_sp = str(sol[xvar].lhs)
                    answer = str(simplify(sol[xvar].rhs))
                    ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                    if var_sp not in my_equations:
                        my_equations[var_sp] = answer
                    used3[var_sp] = True

            except FunctionTimedOut:
                float("jjj")
        except:
            try:
                print(2)
                ind = 0
                not_f = []
                good = []
                for fex in fes:
                    ics = {j_s[ind].subs(tvar, 0): jso[ind]}
                    spc = list(fex.lhs.atoms(Function))
                    spr = list(fex.rhs.atoms(Function))
                    if len(spr) == 1 and spc[0] == spr[0]:
                        # print(3)
                        sol = [dsolve(fex, ics=ics)]
                        good.append([sol[0].lhs, sol[0].rhs])
                        if str(sol[0].lhs) not in used3:
                            var_sp = str(sol[0].lhs)
                            answer = str(simplify(sol[0].rhs))
                            ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                            if var_sp not in my_equations:
                                my_equations[var_sp] = answer
                            used3[var_sp] = True
                    else:
                        # print(4)
                        # print([fex,ics])
                        not_f.append([fex, ics])
                    ind = ind + 1

                # print(good)
                # print(not_f)
                ffprint(["\n\n", "Attempting to evaluate", "\n\n"])
                flen = 2 * len(not_f)
                c_c = 0

                sets = get_sets(r_dict, p_dict)
                fez = []
                # print(slabels,stch_var)
                for svar in sets:
                    fez.append(
                        Eq(sum([c_s[xvar] / stoich(xvar) - c_so[xvar]
                                / stoich(xvar) for xvar in svar]), 0))
                xret = solve(fez, j_s)
                # print(xret)

                while len(not_f) > 0 and c_c < flen:
                    xvar = not_f.pop(0)
                    # print(xvar,100000000000000000000000000000000000)
                    spc = list(xvar[0].lhs.atoms(Function))
                    spr = list(xvar[0].rhs.atoms(Function))
                    for svar in xret:
                        # print(svar not in spc)
                        if svar not in spc:
                            subx = Eq(Symbol("Something_not_needed"),
                                      xret[svar]).rhs.atoms(Function)
                            if spc[0] in subx:
                                # print(xvar[0],xvar[1],111)
                                xvar[0] = xvar[0].subs(svar, xret[svar])
                                # print(xvar[0],xvar[1],222)
                    # print(xvar,100000000000000000000000000000000000)
                    spr = list(xvar[0].rhs.atoms(Function))
                    terms = len(Add.make_args(xvar[0].rhs))
                    if len(spr) == 1 and spc[0] == spr[0]:
                        try:
                            # print(xvar[0],xvar[1],333)
                            sol = [dsolve(xvar[0], ics=xvar[1])]
                            fez.append(Eq(sol[0].lhs[0] - sol[0].rhs, 0))
                            good.append([sol[0].lhs, sol[0].rhs])
                            if str(sol[0].lhs) not in used3:
                                var_sp = str(sol[0].lhs)
                                answer = str(simplify(sol[0].rhs))
                                ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                                if var_sp not in my_equations:
                                    my_equations[var_sp] = answer
                                used3[var_sp] = True
                        except:
                            denom = cs_to_csr(xvar[0].rhs, c_s, cs_r, not_semi)
                            itvar = cs_to_csr(spc[0], c_s, cs_r, not_semi)
                            val = integrate(
                                1 / denom, (itvar,
                                            xvar[1][spc[0].subs(tvar, 0)],
                                            spc[0])) - tvar
                            fez.append(val)
                            sol = solve(val, spc[0])
                            if isinstance(sol, dict):
                                sol = simplify(sol[spc[0]])
                            else:
                                sol = simplify(sol[0])
                            good.append([spc[0], sol])
                            var_sp = str(spc[0])
                            answer = str(sol)
                            ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                            if var_sp not in my_equations:
                                my_equations[var_sp] = answer
                            used3[var_sp] = True
                    elif len(spr) == 2 and terms == 1 \
                        and (len(r_dict) == 1 or len(p_dict) == 1) \
                            and not_semi:
                        subt = Eq(
                            spr[0] - spr[1], jso[j_s.index(spr[0])]
                            - jso[j_s.index(spr[1])])
                        if spr[0] != spc[0]:
                            vari = spr[0]
                        else:
                            vari = spr[1]
                        sol = solve(subt, vari)
                        if isinstance(sol, dict):
                            sol = sol[vari]
                        else:
                            sol = sol[0]
                        subx = Eq(Symbol("Something_not_needed"),
                                  sol).rhs.atoms(Function)
                        if spc[0] in subx:
                            # print(xvar[0],xvar[1],444)
                            xvar[0] = xvar[0].subs(vari, sol)
                            # print(xvar[0],xvar[1],555)
                        not_f.append([xvar[0], xvar[1]])
                    else:
                        not_f.append([xvar[0], xvar[1]])
                    c_c = c_c + 1

                if len(not_f) > 0 and len(used3) < len(j_s):
                    ffprint(
                        ["\n", "The following are difficult to solve",
                         "using dsolve", "\n\n"])
                    for xvar in not_f:
                        ffprint(["\n", xvar[0], "\n\n"])

                    ffprint(
                        ["\n", "The following expression may be wrong,",
                         "please compare with ode_int result :", "\n\n"])
                    c_c = 0
                    while len(not_f) > 0 and c_c < flen:
                        xvar = not_f.pop(0)
                        # print(xvar[0])
                        for svar in xret:
                            if svar not in xvar[0].lhs.atoms(Function):
                                xvar[0] = xvar[0].subs(svar, xret[svar])
                        # print(xvar[0])

                        # for yvar in good:
                            # if yvar[0] not in xvar[0].lhs.atoms(Function):
                                # xvar[0] = xvar[0].subs(yvar[0],yvar[1])
                        try:
                            spc = list(xvar[0].lhs.atoms(Function))
                            spr = list(xvar[0].rhs.atoms(Function))
                            if len(spr) == 1 and spc[0] == spr[0]:
                                spc = spc[0]
                                try:
                                    denom = cs_to_csr(
                                        xvar[0].rhs, c_s, cs_r, not_semi)
                                    itvar = cs_to_csr(spc, c_s, cs_r, not_semi)
                                    val = integrate(
                                        1 / denom, (itvar,
                                                    xvar[1][spc.subs(tvar, 0)],
                                                    spc)) - tvar
                                except:
                                    denom = cs_to_csr(
                                        xvar[0].rhs, c_s, cs_r, False)
                                    itvar = cs_to_csr(spc, c_s, cs_r, False)
                                    val = integrate(
                                        1 / denom, (itvar,
                                                    xvar[1][spc.subs(tvar, 0)],
                                                    spc)) - tvar

                                ffprint(
                                    ["\n", "Not simplified solution", "\n\n"])
                                ffprint(["need to solve for ", spc,
                                         " to simplify :\n", val, "\n\n"])
                                fez.append(val)
                                sol = solve(val, spc)
                                if isinstance(sol, dict):
                                    sol = sol[spc]
                                else:
                                    sol = simplify(sol[0])
                                good.append([spc, sol])
                                # var_sp = str(spc)
                                # answer = str(sol)
                                # ffprint(["\n",var_sp," = ",answer,"\n\n"])
                                # if var_sp not in my_equations:
                                # my_equations[var_sp] = answer
                            else:
                                not_f.append([xvar[0], xvar[1]])
                        except:
                            not_f.append([xvar[0], xvar[1]])
                        c_c = c_c + 1

                    ffprint(["\n", "Attempting to simplify", "\n\n"])
                    # print(fez)
                    sol = solve(fez, j_s)
                    for c_c in sol:
                        rval = Matrix(c_c).subs(equivals)
                        if np.sum(np.array(rval) < 0) > 0:
                            pass
                        else:
                            x_sol = [c_c[xvar] for xvar in range(len(j_s))]
                            ind = 0
                            for xvar in x_sol:
                                var_sp = str(j_s[ind])
                                answer = str(simplify(xvar))
                                ffprint([var_sp, " = ", answer, "\n\n"])
                                if var_sp not in my_equations:
                                    my_equations[var_sp] = answer
                                ind = ind + 1
                            break
            except:
                print(3)
                ind = 0
                not_f = []
                good = []
                for fex in fes:
                    ics = {j_s[ind].subs(tvar, 0): jso[ind]}
                    spc = list(fex.lhs.atoms(Function))
                    spr = list(fex.rhs.atoms(Function))
                    if len(spr) == 1 and spc[0] == spr[0]:
                        # print(3)
                        sol = [dsolve(fex, ics=ics)]
                        good.append([sol[0].lhs, sol[0].rhs])
                        if str(sol[0].lhs) not in used3:
                            var_sp = str(sol[0].lhs)
                            answer = str(sol[0].rhs)
                            ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                            if var_sp not in my_equations:
                                my_equations[var_sp] = answer
                            used3[var_sp] = True
                    else:
                        # print(4)
                        # print([fex,ics])
                        not_f.append([fex, ics])
                    ind = ind + 1

                # print(not_f,1111111111111111111)
                ffprint(["\n\n", "Attempting to evaluate", "\n\n"])
                flen = 2 * len(not_f)
                c_c = 0

                sets = get_sets(r_dict, p_dict)
                fez = []
                # print(slabels,stch_var)

                while len(not_f) > 0 and c_c < flen:
                    xvar = not_f.pop(0)
                    # print(xvar,100000000000000000000000000000000000)
                    spc = list(xvar[0].lhs.atoms(Function))
                    spr = list(xvar[0].rhs.atoms(Function))
                    terms = len(Add.make_args(xvar[0].rhs))
                    right_ok = True
                    for s_s in spr:
                        if s_s in j_s:
                            right_ok = False
                            break
                    # print(xvar,100000000000000000000000000000000000)
                    if (len(spr) == 1 and spc[0] == spr[0]) or right_ok:
                        try:
                            # print(xvar[0],xvar[1],333)
                            sol = [dsolve(xvar[0], ics=xvar[1])]
                            fez.append(Eq(sol[0].lhs[0] - sol[0].rhs, 0))
                            good.append([sol[0].lhs, sol[0].rhs])
                            if str(sol[0].lhs) not in used3:
                                var_sp = str(sol[0].lhs)
                                answer = str(sol[0].rhs)
                                ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                                if var_sp not in my_equations:
                                    my_equations[var_sp] = answer
                                used3[var_sp] = True
                        except:
                            denom = cs_to_csr(xvar[0].rhs, c_s, cs_r, not_semi)
                            itvar = cs_to_csr(spc[0], c_s, cs_r, not_semi)
                            val = integrate(
                                1 / denom, (itvar,
                                            xvar[1][spc[0].subs(tvar, 0)],
                                            spc[0])) - tvar
                            fez.append(val)
                            sol = solve(val, spc[0])
                            if isinstance(sol, dict):
                                sol = simplify(sol[spc[0]])
                            else:
                                sol = simplify(sol[0])
                            good.append([spc[0], sol])
                            var_sp = str(spc[0])
                            answer = str(sol)
                            ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                            if var_sp not in my_equations:
                                my_equations[var_sp] = answer
                            used3[var_sp] = True
                    else:
                        not_f.append([xvar[0], xvar[1]])
                    c_c = c_c + 1

                # print(not_f)
                if len(not_f) > 0 and len(used3) < len(j_s):
                    ffprint(
                        ["\n", "The following are difficult to solve",
                         "using dsolve", "\n\n"])
                    for xvar in not_f:
                        ffprint(["\n", xvar[0], "\n\n"])

                    ffprint(
                        ["\n", "The following expression may be wrong",
                         "please compare with ode_int result :", "\n\n"])
                    c_c = 0
                    while len(not_f) > 0 and c_c < flen:
                        xvar = not_f.pop(0)
                        # print(xvar[0],1111)
                        for yvar in good:
                            if yvar[0] not in xvar[0].lhs.atoms(Function):
                                xvar[0] = xvar[0].subs(yvar[0], yvar[1])
                        # print(xvar[0],2222)
                        try:
                            spc = list(xvar[0].lhs.atoms(Function))
                            spr = list(xvar[0].rhs.atoms(Function))
                            right_ok = True
                            for s_s in spr:
                                if s_s in j_s:
                                    right_ok = False
                                    break

                            # print((len(spr) == 1 and spc[0] == spr[0]) or \
                            # right_ok, (len(spr) == 1 and spc[0] == spr[0]), \
                            # right_ok)
                            if (len(spr) == 1 and spc[0]
                                    == spr[0]) or right_ok:
                                # print(111111111)
                                spc = spc[0]
                                try:
                                    denom = cs_to_csr(
                                        xvar[0].rhs, c_s, cs_r, not_semi)
                                    itvar = cs_to_csr(spc, c_s, cs_r, not_semi)
                                    # print(
                                    #     1/denom,
                                    #     (itvar,xvar[1][spc.subs(tvar,0)],
                                    #      spc),1111)
                                    val = integrate(
                                        1 / denom,
                                        (itvar, xvar[1][spc.subs(tvar, 0)],
                                         spc)) - tvar
                                except:
                                    denom = cs_to_csr(
                                        xvar[0].rhs, c_s, cs_r, False)
                                    itvar = cs_to_csr(spc, c_s, cs_r, False)
                                    # print(1/denom,
                                    #       (itvar,xvar[1][spc.subs(tvar,0)],
                                    #        spc),2222)
                                    val = integrate(
                                        1 / denom,
                                        (itvar, xvar[1][spc.subs(tvar, 0)],
                                         spc)) - tvar

                                ffprint(
                                    ["\n", "Not simplified solution", "\n\n"])
                                ffprint(["need to solve for ", spc,
                                         " to simplify :\n", val, "\n\n"])
                                fez.append(val)
                                sol = solve(val, spc)
                                if isinstance(sol, dict):
                                    sol = sol[spc]
                                else:
                                    sol = simplify(sol[0])
                                good.append([spc, sol])
                                var_sp = str(spc)
                                answer = str(sol)
                                ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                                if var_sp not in my_equations:
                                    my_equations[var_sp] = answer
                            else:
                                not_f.append([xvar[0], xvar[1]])
                        except:
                            not_f.append([xvar[0], xvar[1]])
                        c_c = c_c + 1

                    ffprint(["\n", "Attempting to simplify", "\n\n"])
                    # print(fez)
                    sol = solve(fez, j_s)
                    if isinstance(sol, dict):
                        for c_c in sol:
                            var_sp = str(c_c)
                            answer = str(sol[c_c])
                            ffprint(["\n", var_sp, " = ", answer, "\n\n"])
                            if var_sp not in my_equations:
                                my_equations[var_sp] = answer
                    else:
                        for c_c in sol:
                            rval = Matrix(c_c).subs(equivals)
                            if np.sum(np.array(rval) < 0) > 0:
                                pass
                            else:
                                x_sol = [c_c[xvar] for xvar in range(len(j_s))]
                                ind = 0
                                for xvar in x_sol:
                                    var_sp = str(j_s[ind])
                                    answer = str(simplify(xvar))
                                    ffprint(
                                        ["\n", var_sp, " = ", answer, "\n\n"])
                                    if var_sp not in my_equations:
                                        my_equations[var_sp] = answer
                                    ind = ind + 1
                                break
    # print(slabels)
    return my_equations



"""

             This module is the recalculate_globals module

The purpose of this module is to ensure that global variables holds True
each process for multitrajectory simulation.


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))


import random as rm
import re
from sympy import re as sympy_re
from BioSANS2020.myglobal import mglobals as globals2
from BioSANS2020.math_functs.sbml_math import SBML_FUNCT_DICT


reserve_events_words = {
    "t", "time", "status", "status2", "timer", "finish", "delay", "dtime"}


CONCP = None
RATESP = None
ORASP = None
SIP = None
ACTUALSP = None
CONCP2 = None


def eval2(to_eval):
    """Evaluate a string expression using the SBML_FUNCT_DICT dictionary

    Args:
        to_eval (str): expression

    Returns:
        multitype: evaluated result
    """
    return eval(to_eval, SBML_FUNCT_DICT)


def rate_of(xvar):
    """This is to supplement the SBML rateOf function. The return is the
    instantaneous rate at the current state of the system"""
    xvar = RATESP
    indx = SIP.index(xvar)
    if len(ORASP) < 2:
        return 0
    delt = ORASP[-1] - ORASP[-2]
    rate = (CONCP[-1][indx] - CONCP[-2][indx]) / delt
    return rate


SBML_FUNCT_DICT['rateOf'] = rate_of


def delay_part1(yvar, last):
    """Part of the delay function. See delay function for details
    """
    delt = ORASP[-1] - ORASP[-2]
    inc = int(yvar / delt)
    i = inc
    while abs(abs(last - ORASP[-i]) / yvar - 1.0) > 1.0e-2:
        if last - ORASP[-i] < yvar:
            i = i + 1
        else:
            i = i - 1
    return CONCP[-inc - 1][SIP.index(ACTUALSP)]


def delay_part2(last, delt, yvar):
    """Part of the delay function. See delay function for details
    """
    for yhvar in range(len(globals2.MODIFIED[ACTUALSP])):
        spvar, pfunc = globals2.MODIFIED[ACTUALSP][yhvar]
        spvar = [csp.strip() for csp in spvar]
        if "t" in spvar or "time" in spvar:
            try:
                return pfunc(last + delt - yvar)
            except:
                sprv = []
                for csp, _ in enumerate(spvar):
                    sprv.append(CONCP2[spvar[csp].strip()])
                return pfunc(*sprv)
    return None


def delay(xvar, yvar):
    """This is to supplement the SBML delay function"""
    if ORASP:
        last = ORASP[-1]
        # first = ORASP[0]
        if last - yvar >= 0:
            return delay_part1(yvar, last)
        try:
            # if True:
            if len(ORASP) > 1:
                delt = ORASP[-1] - ORASP[-2]
            else:
                delt = 0
            if ACTUALSP in globals2.MODIFIED:
                val = delay_part2(last, delt, yvar)
                if val:
                    return val
            else:
                return 0
        except:
            xvar = ACTUALSP
            return CONCP[0][SIP.index(xvar)]

    return CONCP[SIP.index(ACTUALSP)]


SBML_FUNCT_DICT['delay'] = delay


def none_to_list(xvar=None):
    """Transform a None variable into a list [] to avoid dangerous
    initial values."""
    if not xvar:
        return []
    return xvar


def apply_rules(conc, yconc, oras=None, spconc=None, slabels=None):
    """This function modifies the concentration base on the rules stated
    in BioSANS topology file.

    Args:
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        yconc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        oras (float, optional): current time point. Defaults to None.
        spconc (list, optional): values of conc at the current time.
            Defaults to None.
        slabels (list, optional): names of components. Defaults to None.
    """    """"""
    global ORASP, CONCP, SIP, ACTUALSP, CONCP2, RATESP
    ORASP = none_to_list(oras)
    CONCP = none_to_list(spconc)
    CONCP2 = none_to_list(conc)
    SIP = none_to_list(slabels)

    tuples_pfunct = []
    for xvar in globals2.MODIFIED:
        for yvar in range(len(globals2.MODIFIED[xvar])):
            spvar, pfunc = globals2.MODIFIED[xvar][yvar]
            ACTUALSP = spvar[0].strip()
            RATESP = spvar[-1].strip()
            try:
                # if True:
                sprv = []
                for csp, _ in enumerate(spvar):
                    sprv.append(conc[spvar[csp].strip()])
                suby = pfunc(*sprv)
                try:
                    suby = sympy_re(suby.evalf())
                except:
                    pass
                if not isinstance(suby, tuple):
                    if suby is not None:
                        yconc[xvar] = suby
                else:
                    tuples_pfunct.append(
                        [suby[1], xvar, suby[0], spvar, pfunc, len(suby)])
            # """
            except:
                suby = pfunc()
                try:
                    suby = sympy_re(suby.evalf())
                except:
                    pass
                if not isinstance(suby, tuple):
                    if suby is not None:
                        yconc[xvar] = suby
                else:
                    tuples_pfunct.append(
                        [suby[1], xvar, suby[0], spvar, pfunc, len(suby)])
            # """

    current_key = None
    update_sp = []
    perst = True
    for val in tuples_pfunct:
        if val[5] == 3:
            perst = False

    while len(tuples_pfunct) > 0:
        tuples_pfunct.sort(key=lambda a: a[0])
        last = 0
        for ihvar, _ in enumerate(tuples_pfunct):
            if tuples_pfunct[last] == tuples_pfunct[ihvar]:
                pass
            else:
                bslice = tuples_pfunct[last:ihvar]
                rm.shuffle(bslice)
                tuples_pfunct[last:ihvar] = bslice
                last = ihvar

        row = tuples_pfunct.pop()
        if current_key is None:
            current_key = row[0]
        if row[2] is not None:
            xvar = row[1]
            yconc[xvar] = row[2]
            if current_key == row[0]:
                update_sp.append(xvar)
            if len(tuples_pfunct) == 0 or current_key != row[0]:
                if perst:
                    for x_o in update_sp:
                        if x_o.split("_")[0] not in reserve_events_words:
                            conc[x_o] = yconc[x_o]
                else:
                    for x_o in update_sp:
                        if row[5] == 3:
                            conc[x_o] = yconc[x_o]
                current_key = row[0]
                update_sp = [xvar]

                for tup in tuples_pfunct:
                    spvar, pfunc = tup[3:5]
                    sprv = []
                    for csp, _ in enumerate(spvar):
                        sprv.append(conc[spvar[csp].strip()])
                    suby = pfunc(*sprv)
                    tup[0] = suby[1]
                    tup[2] = suby[0]

    for xvar in globals2.MODIFIED:
        conc[xvar] = yconc[xvar]


def get_globals(rfile):
    """This function get the global variables from BioSANS topology file
     called rfile.

    Args:
        rfile (str): name of topology file
    """

    globals2.MODIFIED = {}
    globals2.PROP_MODIFIED = {}
    setattr(globals2, 'TCHECK', [])  # TCHECK

    try:
        with open(rfile, "r") as file:
            rows = []
            last = ""
            for row in file:
                row = row.strip()+" "
                if last == "Function_Definitions":
                    if row.strip() != "" and row[0] != "#":
                        exec(row.strip(), SBML_FUNCT_DICT)
                    elif row[0] == "#":
                        last = "#"
                elif last == "#":
                    if row.strip() != "" and row[0] != "@":
                        rows.append(row)
                    elif row.strip() != "" and row[0] != "#":
                        last = "@"
                elif last == "@":
                    if row.strip() != "" and row[0] != "@":
                        cvar = row.split(",")
                        if len(cvar) >= 3:
                            ccv = ",".join(cvar[2:])
                            try:
                                res = re.findall(r"\w{1,3}\(t,.+\)\s", ccv)
                                if res:
                                    try:
                                        rrv = float(str(res[0])
                                                    .split(",")[1]
                                                    .replace(")", ""))
                                    except:
                                        rrv = eval2(str(res[0])
                                                    .split(",")[1]
                                                    .replace(")", ""))
                                    globals2.TCHECK.append(rrv)
                            except:
                                pass
                            cc2 = ccv.split(":")[0].replace(
                                "lambda", "").split(",")

                            delay_pos = ccv.replace(
                                "delay_", "_yaled").find("delay")
                            if delay_pos >= 0:
                                dfirst = ""
                                for ibi in range(delay_pos + 6, len(ccv)):
                                    dvar = ccv[ibi]
                                    if dvar == ",":
                                        break
                                    dfirst = dfirst + dvar
                                dfirst = dfirst.strip()
                                cc2new = [
                                    dfirst] + [
                                        xvar for xvar in cc2
                                        if xvar.strip() != dfirst.strip()]
                                cc2 = cc2new
                                gcc = ccv.split(":")[1]
                                ccv = "lambda " + ",".join(cc2) + ":" + gcc

                            rate_pos = ccv.find("rateOf")
                            if rate_pos >= 0:
                                dfirst = ""
                                for ibi in range(rate_pos + 7, len(ccv)):
                                    dvar = ccv[ibi]
                                    if dvar == ")":
                                        break
                                    dfirst = dfirst + dvar
                                dfirst = dfirst.strip()
                                cc2new = [
                                    xvar for xvar
                                    in cc2 if xvar.strip() != dfirst.strip()
                                ] + [dfirst]
                                cc2 = cc2new
                                gcc = ccv.split(":")[1]
                                ccv = "lambda " + ",".join(cc2) + ":" + gcc

                            if cvar[0].strip() not in globals2.MODIFIED:
                                globals2.MODIFIED[cvar[0].strip()] = [
                                    [cc2, eval2(ccv)]]
                            else:
                                globals2.MODIFIED[cvar[0].strip()].append(
                                    [cc2, eval2(ccv)])
                elif row[0] == "#":
                    last = "#"
                    ggv = row.split(",")[1:]
                    try:
                        for xvar in ggv:
                            xxvar = xvar.split("=")
                            globals2.SETTINGS[xxvar[0].strip()
                                              ] = xxvar[1].strip()
                    except:
                        pass
                elif row[0] == "@":
                    last = "@"
                elif row.strip().upper() == "FUNCTION_DEFINITIONS:":
                    last = "Function_Definitions"
            file.close()

        rxn = len(rows)
        for ihvar in range(rxn):
            col_row = rows[ihvar].split(":::::")
            row = col_row[0].strip().split(",")
            if len(col_row) > 1:
                krow = col_row[1].strip().split(":::")
                if len(krow) == 2:
                    cc2 = krow[0].split(":")[0].replace(
                        "lambda", "").split(",")
                    cc3 = krow[1].split(":")[0].replace(
                        "lambda", "").split(",")
                    globals2.PROP_MODIFIED["Prop_" + str(ihvar)] = \
                        [(cc2, eval2(krow[0])), (cc3, eval2(krow[1]))]
                else:
                    cc2 = krow[0].split(":")[0].replace(
                        "lambda", "").split(",")
                    globals2.PROP_MODIFIED["Prop_" + str(ihvar)] = \
                        [(cc2, eval2(krow[0]))]

        globals2.TCHECK.sort()
    except:
        pass



"""

                 This module is the gillespie_ssa module

This module propagates the stochastic simulation algorithm also known as
Gillespie.

"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import numpy as np
from BioSANS2020.propagation.propensity import propensity_vec
from BioSANS2020.propagation.recalculate_globals \
    import get_globals, apply_rules, reserve_events_words
from BioSANS2020.myglobal import mglobals as globals2


def gillespie_ssa(tvar, sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                  rand_seed, implicit=False, rfile=""):
    """[summary]

    Args:
        tvar (list): time stamp of simulation
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        rand_seed (float): random  seed value picked  at random for each
            trajectory. They have been sampled from the calling program.
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        rfile (string, optional): name of topology file where some
            parameters or components are negative indicating  they  have
            to be estimated. Defaults to None.

    Returns:
        tuple: (time, trajectories)
    """
    get_globals(rfile)
    tmax = tvar[-1]
    np.random.seed(int(rand_seed * 100))
    tnew = []
    stch_var = stch_var.T
    all_sp = list(sp_comp.keys())  # [z for z in sp_comp]
    spc = [z for z in sp_comp if z not in reserve_events_words]
    spc2 = [z for z in sp_comp if z in reserve_events_words]
    concz = {xvar: conc[xvar] for xvar in conc}
    yconc = {xvar: conc[xvar] for xvar in conc}
    apply_rules(concz, yconc)
    update_sp = [all_sp.index(z) for z in spc]

    zlist = [[concz[z] for z in spc]]
    t_c = 0
    tnew.append(t_c)
    if not implicit:
        while t_c < tmax:
            prop_flux = propensity_vec(ks_dict, concz, r_dict, p_dict)
            alp = np.sum(prop_flux)
            r_1 = np.random.uniform()
            while r_1 == 0:
                r_1 = np.random.uniform()
            p_r = np.cumsum([d / alp for d in prop_flux])
            d_t = (1 / alp) * (np.log(1 / r_1))
            if np.isnan(d_t) or np.isinf(d_t):
                break
            r_2 = np.random.uniform()
            for i, _ in enumerate(p_r):
                if r_2 <= p_r[i]:
                    all_pos = True
                    for xvar, _ in enumerate(spc):
                        holder = zlist[-1][xvar] + stch_var[i][update_sp[xvar]]
                        if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                            concz[spc[xvar]] = holder
                        else:
                            all_pos = True
                            break
                    if all_pos:
                        for xvar, _ in enumerate(spc2):
                            concz[spc2[xvar]] = concz[spc2[xvar]] + d_t
                        apply_rules(concz, yconc)
                        zlist.append([concz[xvar] for xvar in spc])
                        t_c = t_c + d_t
                        tnew.append(t_c)
                    else:
                        for xvar, _ in enumerate(spc):
                            concz[spc[xvar]] = zlist[-1][xvar]
                    break
    else:
        z_c = []
        tindex = 0
        index = 0
        tchlen = len(globals2.TCHECK)
        while t_c < tmax:
            prop_flux = propensity_vec(ks_dict, concz, r_dict, p_dict)
            alp = np.sum(prop_flux)
            r_1 = np.random.uniform()
            while r_1 == 0:
                r_1 = np.random.uniform()
            p_r = np.cumsum([d / alp for d in prop_flux])
            d_t = (1 / alp) * (np.log(1 / r_1))

            if index != tchlen:
                if t_c + d_t >= globals2.TCHECK[index]:
                    d_t = globals2.TCHECK[index] - t_c
                    index = index + 1

            if np.isnan(d_t) or np.isinf(d_t):
                z_c.append(zlist[-1])
                while tvar[tindex] != tmax:
                    z_c.append(zlist[-1])
                    tindex = tindex + 1
                break
            r_2 = np.random.uniform()
            for i, _ in enumerate(p_r):
                if r_2 <= p_r[i]:
                    all_pos = True
                    for xvar, _ in enumerate(spc):
                        holder = zlist[-1][xvar] + stch_var[i][update_sp[xvar]]
                        if holder >= 0 or spc[xvar] in globals2.MODIFIED:
                            concz[spc[xvar]] = holder
                        else:
                            all_pos = False
                            break
                    if all_pos:
                        for xvar, _ in enumerate(spc2):
                            concz[spc2[xvar]] = concz[spc2[xvar]] + d_t

                        t_c = t_c + d_t
                        if "t" in spc2:
                            concz["t"] = t_c
                        elif "time" in spc2:
                            concz["time"] = t_c
                        else:
                            pass

                        apply_rules(concz, yconc)
                        zlist.append([concz[xvar] for xvar in spc])
                        # t_c = t_c + d_t
                        try:
                            if t_c == tvar[tindex]:
                                z_c.append(zlist[-1])
                                tindex = tindex + 1
                            else:
                                while t_c > tvar[tindex]:
                                    z_c.append(zlist[-2])
                                    tindex = tindex + 1
                        except:
                            pass
                        tnew.append(t_c)
                    else:
                        for xvar, _ in enumerate(spc):
                            concz[spc[xvar]] = zlist[-1][xvar]
                    break
        tnew = tvar
        zlist = z_c
    return (tnew, np.array(zlist))
